!*******************************************************************
!*                           PENAUX                                *
!*                                                                 *
!* Short description:                                              *
!*   Miscelaneous routines that serve to initialize and monitor a  *
!*   PENELOPE main program.                                        *
!*******************************************************************


      subroutine iniconfig(restartf)
!*******************************************************************
!*    Reads simulation configuration.                              *
!*                                                                 *
!*    Output:                                                      *
!*      restartf -> Restart file ID.                               *
!*      Other output is through common blocks.                     *
!*******************************************************************
      implicit none
      integer restartf

      integer*4 seed1,seed2
      common/rseed/seed1,seed2
      real*8 nhist,time0,atime,refresh,ncalls,lastime
      common /ctrsim/ nhist,time0,atime,refresh,ncalls,lastime
      character*80 dumpfilen
      integer restartfile,dumpfile
      common /dump/ restartfile,dumpfile,dumpfilen
      character*80 buffer
      character*(*) secid,eos
      parameter (secid='[SECTION CONFIG v.2012-06-02]')
      parameter (eos='[END OF CONFIG SECTION]')
      integer ufile,finduf,error
      real*8 maxn
      parameter (maxn=1.0d15)

      write(*,*) ''
      write(*,'(a)')
     & '>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>'
      call getline(buffer)
      if (index(buffer,secid).eq.0) then
        write(*,'(a)') 'iniconfig:ERROR: incorrect section header;'
        write(*,'(a,a)') '  expecting to find: ',secid
        write(*,'(a,a)') '  found instead:     ',buffer
        stop
      endif
      write(*,'(a)') secid

      write(*,'(a)') 'No. of histories requested: '
      read(*,*) nhist
      write(*,'(es12.5)') nhist
      if (nhist.gt.maxn) then
        write(*,'(a,es12.5)')
     &    'iniconfig:ERROR: No. of requested histories exceeds ',maxn
        stop
      endif

      write(*,'(a)') 'Allotted time (s):'
      read(*,*) atime
      write(*,'(es12.5)') atime

      write(*,'(a)') 'Update interval:'
      read(*,*) refresh
      write(*,'(es12.5)') refresh
      if (refresh.gt.5.0e4) then
        write(*,'(a)') 'iniconfig:ERROR: update time must be < 50000'
        stop
      endif

      write(*,'(a)') 'Random seeds:'
      read(*,*) seed1,seed2
      write(*,'(2(1x,i0))') seed1,seed2
      restartfile = -1       ! Indicates no restart requested
      if (seed1.eq.0.and.seed2.eq.0)  then
        continue             ! External seeds file requested, to be processed later
      else if (seed1.eq.-1.and.seed2.eq.-1) then
        restartfile = 1      ! Restart sim from a previous dump file
      else if (seed1.lt.1.or.seed2.lt.1) then
        write(*,'(a)') 'iniconfig:ERROR: Invalid RNG seeds.'
        stop
      endif

      read(*,'(a80)') buffer                 ! Read seeds filename
      if (seed1.eq.0.and.seed2.eq.0) then
        buffer = adjustl(buffer)             ! Remove leading blanks
        buffer = buffer(1:scan(buffer,' '))  ! Clip at 1st blank
        write(*,'(a)') 'Reading seeds from external file named:'
        write(*,'(1x,a)') buffer
        ufile = finduf()
        open(ufile,file=buffer,status='old',iostat=error)
        if (error.ne.0) then
          write(*,'(a)') 'iniconfig:ERROR: unable to open seeds file.'
          stop
        endif
        read(ufile,*) seed1,seed2
        close(ufile)
        write(*,'(a)') 'Random seeds:'
        write(*,'(2(1x,i0))') seed1,seed2
        if (seed1.lt.1.or.seed2.lt.1) then
          write(*,'(a)') 'iniconfig:ERROR: Invalid RNG seeds.'
          stop
        endif
      endif

      read(*,'(a80)') buffer ! Restart filename
      if (restartfile.eq.1) then
        write(*,'(a)')
     &    '*** Restarting simulation from a previous run***'
        buffer = adjustl(buffer)
        buffer = buffer(1:scan(buffer,' '))
        write(*,'(a)') 'Opening restart file:'
        write(*,'(1x,a)') buffer
        restartfile = finduf()
        open(restartfile,file=buffer,status='old',access='stream',
     &       iostat=error)
        if (error.ne.0) then
          write(*,'(a)')
     &      'iniconfig:ERROR: unable to open restart file.'
          stop
        endif
      endif
      restartf = restartfile  ! Output file unit to inform main program

      read(*,'(a80)') buffer ! Output dump filename
      buffer = adjustl(buffer)
      dumpfilen = buffer(1:scan(buffer,' '))
      dumpfile = -1          ! Indicates no dump file requested
      if (dumpfilen.ne.'-') then
        write(*,'(a)') 'Simulation dump requested. Dump file name:'
        write(*,'(1x,a)') dumpfilen
        dumpfile = 1
      endif

      ncalls = 0

      read(*,'(a80)') buffer
      if (index(buffer,eos).eq.0) then
        write(*,'(a)')
     &    'iniconfig:ERROR: End-Of-Section mark not found'
        write(*,'(a,a)') '  expecting to find: ',eos
        write(*,'(a,a)') '  found instead:     ',buffer
        stop
      endif
      write(*,'(a)') '>>>> Config initialization finished >>>>'
      end


      subroutine inigeo(nmat)
!*******************************************************************
!*    Reads geometry section and initializes PENGEOM and penVox.   *
!*                                                                 *
!*    Output:                                                      *
!*      nmat -> largest material index in GEO & VOX files          *
!*******************************************************************
      implicit none
      integer*4 nmat

      logical isQuad
      common /geoquad/ isQuad
      character*80 buffer
      character*(*) secid,eos
      parameter (secid=
     &'[SECTION PENGEOM+PENVOX v.2009-06-15]')
      parameter (eos='[END OF GEO SECTION]')
      integer ufile,finduf,error
      integer*4 maxmat,nmatvox,nmatquad,nbody
      parameter (maxmat=10)
      real*8 parinp(1)

      write(*,*) ''
      write(*,'(a)')
     & '>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>'
      call getline(buffer)
      if (index(buffer,secid).eq.0) then
        write(*,'(a)') 'inigeo:ERROR: incorrect section header;'
        write(*,'(a,a)') '  expecting to find: ',secid
        write(*,'(a,a)') '  found instead:     ',buffer
        stop
      endif
      write(*,'(a)') secid

      ! Read quadrics file, if there is one:
      write(*,*) ''
      write(*,'(a)')
     &  '>>>> Initializing quadric geometry >>>>'
      read(*,'(a80)') buffer ! Quadrics filename
      buffer = adjustl(buffer)
      buffer = buffer(1:scan(buffer,' ')) ! Clip at 1st blank
      if (buffer.ne.'-') then
        isQuad = .true.
        write(*,'(a)') 'Opening quadric geometry file:'
        write(*,'(1x,a)') buffer
      else  ! No quadrics
        isQuad = .false.
        buffer = 'default.tmp.geo'
        write(*,'(a,a,a)')
     &    '(No quadrics; an empty quadric geometry '//
     &    'will be automatically created in file ',
     &    trim(buffer),')'
        call autogeo(buffer)
      endif

      ! Init quadric geom (PENGEOM):
      ufile = finduf()
      open(ufile,file=buffer,status='old',iostat=error)
      if (error.ne.0) then
        write(*,'(a)') 'inigeo:ERROR: unable to open quadrics file'
        stop
      endif
      write(*,'(a)') 'Now calling PENGEOM...'
      write(*,*) ''
      call geomin(parinp,1,nmatquad,nbody,ufile,6)
      close(ufile)
      write(*,*) ''
      write(*,'(a)')
     &  '>>>> Quadric geometry initialization finished >>>>'

      ! Init voxel geom (penVox), if there is one:
      call inivox(nmatvox)
      ! Note: some additional init procedures are performed by VDD tally

      ! Set no. of materials in geometry files:
      nmat = max(nmatvox,nmatquad)
      write(*,*) ''
      write(*,'(a)') 'No. of materials referenced in the geometry:'
      write(*,'(1x,i0)') nmat
      if (nmat.gt.maxmat) then
        write(*,'(a)')
     &    'inigeo:ERROR: too many materials; '//
     &    'enlarge MAXMAT in Fortran code and recompile.'
        stop
      else if (nmat.eq.0) then
        write(*,'(a)') 'inigeo:ERROR: no geometry defined.'
        stop
      endif

      ! Check section integrity:
      read(*,'(a80)') buffer
      if (index(buffer,eos).eq.0) then
        write(*,'(a)') 'inigeo:ERROR: End-Of-Section mark not found'
        write(*,'(a,a)') '  expecting to find: ',eos
        write(*,'(a,a)') '  found instead:     ',buffer
        stop
      endif
      write(*,*) ''
      write(*,'(a)') '>>>> Geometry initialization finished >>>>'
      end


      subroutine inipen(emax,nmatgeo)
!*******************************************************************
!*    Reads simulation parameters and initializes PENELOPE.        *
!*                                                                 *
!*    Input:                                                       *
!*      emax -> max source energy (eV)                             *
!*      nmatgeo -> max number of materials in GEO & VOX files      *
!*******************************************************************
      implicit none
      integer*4 nmatgeo
      real*8 emax

      integer*4 maxmat
      parameter (maxmat=10)
      real*8 eabs,c1,c2,wcc,wcr
      common/csimpa/eabs(3,maxmat),c1(maxmat),c2(maxmat),
     &              wcc(maxmat),wcr(maxmat)
      real*8 dsmaxval
      common /ctrsi1/ dsmaxval(maxmat)
      character*80 buffer
      character*300 lbuffer  ! Longer buffer for transport param table
      character*20 matfile(maxmat),filename
      character*(*) secid,eos
      parameter (secid=
     &'[SECTION PENELOPE v.2009-10-01]')
      parameter (eos='[END OF PEN SECTION]')
      integer i,error,imat
      integer*4 nmat
      real*8 infty
      parameter (infty=1.0d30)

      write(*,*) ''
      write(*,'(a)')
     & '>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>'
      call getline(buffer)
      if (index(buffer,secid).eq.0) then
        write(*,'(a)') 'inipen:ERROR: incorrect section header;'
        write(*,'(a,a)') '  expecting to find: ',secid
        write(*,'(a,a)') '  found instead:     ',buffer
        stop
      endif
      write(*,'(a)') secid

      write(*,*) ''
      read(*,'(a300)') lbuffer  ! Read header
      write(*,'(a)') 'MAT#  FILENAME             EABS(e-)  EABS(ph)'//
     &  '  EABS(e+)  C1        C2        WCC       WCR       DSMAX'

      nmat = 0
      do i=1,maxmat
        read(*,'(a300)') lbuffer

        ! Read mat index:
        read(lbuffer,*) imat
        if (imat.eq.0) exit  ! End-of-list
        if (imat.ne.i) then
          write(*,'(a,i0,a,i0)')
     &      'inipen:ERROR: Invalid MAT index; found ',imat,
     &      ', should have been ',i
          stop
        endif
        lbuffer = adjustl(lbuffer) ! Remove leading blanks
        lbuffer = lbuffer(scan(lbuffer,' '):) ! Remove 1st datum

        ! Read mat file name:
        lbuffer = adjustl(lbuffer)
        read(lbuffer,'(a20)') filename
        filename = filename(1:scan(filename,' ')) ! Clip at 1st blank
        if (len_trim(filename).eq.0) exit  ! Empty MAT field, list is over
        matfile(i) = filename
        nmat = i
        lbuffer = adjustl(lbuffer) ! Remove leading blanks
        lbuffer = lbuffer(scan(lbuffer,' '):) ! Remove 1st datum

        ! Read transport parameters:
        eabs(1,i) = -1.0
        read(lbuffer,*,iostat=error) eabs(1,i),eabs(2,i),eabs(3,i),
     &                  c1(i),c2(i),wcc(i),wcr(i),dsmaxval(i)
        if (error.ne.0) then
          if (eabs(1,i).ne.-1.0) then
            write(*,'(a,i0)')
     &       'inipen:ERROR: incomplete list of parameters for MAT ',i
            stop
          endif

          ! Set parameters automatically:
          eabs(1,i) = min(max(emax*1.0d-2,50.0d0),1.0d6)
          eabs(2,i) = min(max(emax*1.0d-3,50.0d0),1.0d6)
          eabs(3,i) = eabs(1,i)
          c1(i)     = 0.1d0
          c2(i)     = 0.1d0
          wcc(i)    = min(emax*1.0d-2,eabs(1,i))
          wcr(i)    = min(emax*1.0d-3,eabs(2,i))
          dsmaxval(i)  = +infty
        endif

        ! Write data:
        write(*,'(i2,4x,a20,8(1x,es9.3))')
     &    i,filename,eabs(1,i),eabs(2,i),
     &    eabs(3,i),c1(i),c2(i),wcc(i),wcr(i),dsmaxval(i)
      enddo

      if (nmatgeo.gt.nmat) then
        write(*,'(a)')
     &    'inipen:ERROR: There are more materials declared in the'
        write(*,'(a)')
     &    '   geometry file than defined in the config file'
        stop
      endif

      ! Init PENELOPE kernel:
      write(*,*) ''
      write(*,'(a)') 'Now calling PEINIT...'
      write(*,*) ''
      call peinit(emax*1.01d0,nmat,6,1,matfile)  ! Add 1% to EMAX to allow perfect absorbents

      read(*,'(a80)') buffer
      if (index(buffer,eos).eq.0) then
        write(*,'(a)') 'inipen:ERROR: End-Of-Section mark not found'
        write(*,'(a,a)') '  expecting to find: ',eos
        write(*,'(a,a)') '  found instead:     ',buffer
        stop
      endif
      write(*,*) ''
      write(*,'(a)') '>>>> PENELOPE initialization finished >>>>'
      end


      logical function endsim(n)
!*******************************************************************
!*    Determines whether the simulation is done or not and writes  *
!*    progress reports.                                            *
!*                                                                 *
!*    Input:                                                       *
!*      n -> current history no.                                   *
!*    Output:                                                      *
!*      .true. if done                                             *
!*******************************************************************
      implicit none
      real*8 n

      real*8 nhist,time0,atime,refresh,ncalls,lastime
      common /ctrsim/ nhist,time0,atime,refresh,ncalls,lastime
      logical needfresh
      integer unc
      real*8 realtime,cputime,nowreal,nowcpu

      nowreal = realtime() ! Ensures one call per day at least
      nowcpu  = cputime()
      endsim  = .false.

      ! Check no. of histories:
      if (n-nhist.gt.-0.5d0) endsim = .true.

      ! Check time:
      if (atime.gt.0.0) then
        if (nowreal.gt.atime) endsim = .true.  ! Real time
      else
        if (nowcpu.gt.-atime) endsim = .true.  ! User (CPU) time
      endif

      if (endsim) return

      ! Check if an update is needed:
      needfresh = .false.
      if (refresh.gt.0.0) then
        if (nowreal-lastime.gt.refresh) then
          lastime = nowreal
          needfresh = .true.
        endif
      else
        ncalls = ncalls+1
        if (ncalls.ge.-refresh) then
          ncalls = 0
          needfresh = .true.
        endif
      endif

      ! Update:
      if (needfresh) then
        call comand(n)
        call tallyreport(n,nowcpu,unc)  ! Writes progress report
        endsim = unc.gt.1               ! Required accuracy attained?
      endif
      end


      subroutine noverflow(n)
!*******************************************************************
!*    Determines whether the number of histories overflows the     *
!*    real*8 counter; if it does, if forces the simulation to stop.*
!*                                                                 *
!*    Input:                                                       *
!*      n -> current history no.                                   *
!*******************************************************************
      implicit none
      real*8 n

      real*8 maxn
      parameter (maxn=1.0d15)

      if (n.gt.maxn) then
        write(*,*) ''
        write(*,'(a)') '***************************************** '
        write(*,'(a,f20.0)')
     &  'noverflow:WARNING: No. of histories too large for'
        write(*,'(a)')
     &  '  penEasy real*8 counters to handle.'
        write(*,'(a)')
     &  '  Forcing the simulation to stop after current history.'
        write(*,'(a)') '***************************************** '
        write(*,*) ''
        call simstop
      endif
      end


      subroutine simstop
!*******************************************************************
!*    Forces the simulation to stop after completion of current    *
!*    history by re-setting the no. of histories to simulate to    *
!*    zero.                                                        *
!*******************************************************************
      implicit none
      real*8 nhist,time0,atime,refresh,ncalls,lastime
      common /ctrsim/ nhist,time0,atime,refresh,ncalls,lastime

      nhist = 0
      end


      real*8 function dsmax()
!*******************************************************************
!*    Maximum step length as passed to JUMP; this function is an   *
!*    interface to common /CTRSI1/                                 *
!*                                                                 *
!*    Input:                                                       *
!*      /TRACK/                                                    *
!*******************************************************************
      implicit none
      integer*4 kpar,ibody,mat,ilb
      real*8 e,x,y,z,u,v,w,wght
      common/track/e,x,y,z,u,v,w,wght,kpar,ibody,mat,ilb(5)
      integer*4 maxmat
      parameter (maxmat=10)
      real*8 dsmaxval
      common /ctrsi1/ dsmaxval(maxmat)

      dsmax = dsmaxval(mat)
      end


      subroutine killpart
!*******************************************************************
!*    Forces current particle to be killed (i.e. absorbed without  *
!*    further actions, e.g., create two annihilation photons for a *
!*    e+) at the next call to ABSORB().                            *
!*******************************************************************
      implicit none
      integer*4 kpar,ibody,mat,ilb
      real*8 e,x,y,z,u,v,w,wght
      common/track/e,x,y,z,u,v,w,wght,kpar,ibody,mat,ilb(5)

      wght = 0.0d0
      end


      logical function absorb()
!*******************************************************************
!*    Checks whether or not a particle is absorbed given its       *
!*    current energy.                                              *
!*                                                                 *
!*    Input:                                                       *
!*      /TRACK/                                                    *
!*    Output:                                                      *
!*      .true. if absorbed, .false. else                           *
!*    Comments:                                                    *
!*      -> Tallies deposited energy when necessary; to do so it    *
!*         calls TALLY.                                            *
!*      -> Absorbs particles above 1 GeV; these can be created by  *
!*         SOURCE or be the result of the annihilation of a e+.    *
!*      -> This function is well suited to implement range         *
!*         rejection by making the returned value dependent        *
!*         on the particle's position, material, energy, etc.      *
!*         Notice that in order to preserve the calculation of     *
!*         the fluence unbiased an electron or positron should     *
!*         *never* be rejected *inside* the detector.              *
!*******************************************************************
      implicit none
      integer*4 kpar,ibody,mat,ilb
      real*8 e,x,y,z,u,v,w,wght
      common/track/e,x,y,z,u,v,w,wght,kpar,ibody,mat,ilb(5)
      integer*4 maxmat
      parameter (maxmat=10)
      real*8 eabs,c1,c2,wcc,wcr
      common/csimpa/eabs(3,maxmat),c1(maxmat),c2(maxmat),
     &              wcc(maxmat),wcr(maxmat)
      real*8 mc2,twomc2,penemax
      parameter (mc2=5.10998918d5,twomc2=2.0d0*mc2,penemax=1.0d9)  ! Electron rest energy

      if (wght.eq.0.0d0) then  ! Killed
        call tally(5,0.0d0)
        absorb = .true.
        return
      endif

      if (mat.eq.0) then  ! In vacuum
        absorb = .true.
        return
      endif

      if (e.gt.eabs(kpar,mat).and.e.lt.penemax) then
        absorb = .false.
        return
      endif

      absorb = .true.
      call tally(-98,e)         ! Tallies remaining kinetic energy
      if (kpar.eq.3.and.e.gt.0.0d0) then  ! Precaution in case of positron
        call panar              ! Puts two annihilation photons in the stack
        call tally(-97,twomc2)  ! Tally the extra 2mc^2
      endif
      if (e.ge.penemax) then
        write(*,*) ''
        write(*,'(a)') '***************'
        write(*,'(a,es12.5,a)')
     &  'absorb:WARNING: particle with energy out of range: ',e,' eV'
        write(*,'(a)') '  (locally absorbed to prevent program abort)'
        write(*,'(a)') '***************'
        write(*,*) ''
      endif
      end


      subroutine comand(n)
!*******************************************************************
!*    Reads and executes commands from an external file, allowing  *
!*    in-flight steering of the simulation.                        *
!*                                                                 *
!*    Input:                                                       *
!*      n -> current no. of histories already simulated.           *
!*      -> Command is read from external file command.in           *
!*    Output:                                                      *
!*      -> Command is executed and file command.in reset to default*
!*******************************************************************
      implicit none
      real*8 n

      real*8 nhist,time0,atime,refresh,ncalls,lastime
      common /ctrsim/ nhist,time0,atime,refresh,ncalls,lastime
      character*80 buffer
      logical ffound
      integer com,in,finduf,error
      real*8 param,maxn
      parameter (maxn=1.0d15)

      inquire(file='command.in',exist=ffound)
      if (.not.ffound) return  ! File not found
      in = finduf()
      open(in,file='command.in',iostat=error)
      if (error.ne.0) return

      ! Read command file:
      iter: do
        read(in,*,iostat=error) com
        if (error.ne.0.or.com.eq.0) exit iter
        write(*,*) ''
        write(*,'(a,f18.0)')
     &   'comand:WARNING: command received when history number was: ',
     &    n
        select case (com)
        case (1)
          read(in,*,iostat=error) param
          if (error.ne.0) exit iter
          nhist = param
          if (nhist.gt.maxn) then
            write(*,*) ''
            write(*,'(a,f18.0)')
     &      'comand:WARNING: requested max No. of histories exceeds ',
     &        maxn
            nhist = maxn
          endif
          write(*,'(a,f18.0)') '  max No. of histories reset to: ',
     &      nhist
        case (2)
          read(in,*,iostat=error) param
          if (error.ne.0) exit iter
          atime = param
          write(*,'(a,es12.5)') '  max time reset to: ',atime
        case (3)
          read(in,*,iostat=error) param
          if (error.ne.0) exit iter
          refresh = param
          write(*,'(a,es12.5)') '  update interval reset to: ',refresh
        case (9)
          read(in,'(a80)',iostat=error) buffer
          if (error.ne.0) exit iter
          write(*,'(a)') '  message from command.in follows:'
          write(*,'(a80)') buffer
        case default
          write(*,'(a,i0,a)')
     &     '  unknown command: ',com,'; remaining commands ignored.'
          exit iter
        end select
      enddo iter
      close(in,iostat=error)

      ! Reset command file:
      open(in,file='command.in',iostat=error)
      if (error.ne.0) return
      write(in,'(a)',iostat=error)
     &  ' 0   <- write here the command code'
      write(in,'(a)',iostat=error)
     &  '     <- write here the command parameter, if any'
      write(in,'(a)',iostat=error)
     &  '     <- repeat previous lines as many times as needed'
      write(in,'(a)',iostat=error)
     &  ''
      write(in,'(a)',iostat=error)
     &  '>>>> END OF INPUT >>>>'
      write(in,'(a)',iostat=error)
     &  ''
      write(in,'(a)',iostat=error)
     &  'Code  Parameter  What it does'
      write(in,'(a)',iostat=error)
     &  '-----------------------------'
      write(in,'(a)',iostat=error)
     &  '0     None       Nothing, keep going'
      write(in,'(a)',iostat=error)
     &  '1     N          Reset No. of histories to N'
      write(in,'(a)',iostat=error)
     &  '2     t          Reset simulation time to t'
      write(in,'(a)',iostat=error)
     &  '3     U          Reset update interval to U'
      write(in,'(a)',iostat=error)
     &  '9     <string>   Write <string> (<80 chars) to output'
      write(in,'(a)',iostat=error)
     &  ''
      write(in,'(a)',iostat=error)
     &  '>>>> END OF FILE >>>>'
      write(in,'(a)',iostat=error)
     &  ''
      close(in,iostat=error)
      end


      integer function finduf()
!*******************************************************************
!*    Finds a valid (i.e. unused) file unit.                       *
!*******************************************************************
      implicit none
      logical used
      integer maxuf
      parameter (maxuf=17)

      finduf = 6
      do
        finduf = finduf+1
        if (finduf.gt.maxuf) then
          write(*,*) ''
          write(*,'(a)') 'finduf:ERROR: Unable to find a valid unit'
          stop
        endif
        inquire(finduf,opened=used)
        if (.not.used) return
      enddo
      end


      subroutine getline(buffer)
!*******************************************************************
!*    Reads a new line from the keyboard. The line is returned     *
!*    only if it is not blank or a comment line.                   *
!*                                                                 *
!*    Output:                                                      *
!*      buffer -> line read.                                       *
!*    Comments:                                                    *
!*      -> comment lines start with a '#'                          *
!*******************************************************************
      implicit none
      character*80 buffer
      integer i

      do
        read(*,'(a80)') buffer
        if (buffer(1:1).eq.'#') cycle  ! A comment line
        do i=1,len(buffer)
          if (buffer(i:i).ne.' ') return
        enddo
      enddo
      end


      integer function seeki(x,xc,n)
!*******************************************************************
!*    Finds the interval (x(i),x(i+1)] containing the value xc.    *
!*                                                                 *
!*    Input:                                                       *
!*      x(1..n) -> data array                                      *
!*      xc -> point to be located                                  *
!*      n -> no. of data points                                    *
!*    Output:                                                      *
!*      index i of the semiopen interval where xc lies             *
!*    Comments:                                                    *
!*      -> If xc=x(1) then i=1 is returned.                        *
!*      -> If xc is outside the closed interval [x(1),x(n)]  the   *
!*         execution is aborted.                                   *
!*******************************************************************
      implicit none
      integer n
      real*8 xc,x(n)

      integer itop,imid

      if(xc.gt.x(n)) then
        write(*,*) 'seeki error: value outside range, xc>x(n):'
        write(*,*) xc,x(n)
        stop
      endif
      if(xc.lt.x(1)) then
        write(*,*) 'seeki error: value outside range, xc<x(1):'
        write(*,*) xc,x(1)
        stop
      endif

      seeki = 1
      itop = n
      do
        imid = (seeki+itop)/2
        if(xc.gt.x(imid)) then
          seeki = imid
        else
          itop = imid
        endif
        if(itop-seeki.le.1) exit
      enddo
      end


      subroutine rotation(omega,theta,phi,rot)
!*******************************************************************
!*    Computes the rotation matrix determined by the three Euler   *
!*    angles as defined in the PENELOPE manual.                    *
!*                                                                 *
!*    Input:                                                       *
!*      omega -> rotation angle around the z axis (rad)            *
!*      theta -> rotation angle around the y axis (rad)            *
!*      phi   -> rotation angle around the z axis (rad)            *
!*    Output:                                                      *
!*      rot -> 3x3 matrix of the 3D rotation                       *
!*******************************************************************
      implicit none
      real*8 omega,theta,phi,rot(3,3)

      real*8 somega,comega,stheta,ctheta,sphi,cphi

      somega = sin(omega)
      comega = cos(omega)
      stheta = sin(theta)
      ctheta = cos(theta)
      sphi   = sin(phi)
      cphi   = cos(phi)
      rot(1,1) = cphi*ctheta*comega-sphi*somega
      rot(1,2) = -cphi*ctheta*somega-sphi*comega
      rot(1,3) = cphi*stheta
      rot(2,1) = sphi*ctheta*comega+cphi*somega
      rot(2,2) = -sphi*ctheta*somega+cphi*comega
      rot(2,3) = sphi*stheta
      rot(3,1) = -stheta*comega
      rot(3,2) = stheta*somega
      rot(3,3) = ctheta
      end


!>>>> End Of File >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
