!*******************************************************************
!*                          TALLY                                  *
!*                     FLUENCE SPECTRUM                            *
!*                  (TRACK LENGTH ESTIMATOR)                       *
!*                                                                 *
!*                                                                 *
!* Short description:                                              *
!*   Tallies track lengths in order to estimate the fluence spectra*
!*   of each type of particle in a given material.                 *
!*                                                                 *
!* Comments:                                                       *
!*   Use of KNOCKX instead of KNOCK in MAIN is required.           *
!*******************************************************************


      subroutine FTLtally(mode,arg)
!*******************************************************************
!*    Input:                                                       *
!*      mode -> identifies the state of the calling procedure      *
!*      arg -> different meanings for different modes; see MAIN    *
!*******************************************************************
      implicit none
      integer mode
      real*8 arg

      integer*4 kpar,ibody,mat,ilb
      real*8 e,x,y,z,u,v,w,wght
      common/track/e,x,y,z,u,v,w,wght,kpar,ibody,mat,ilb(5)
      integer maxmat
      real*8 eabs,c1,c2,wcc,wcr
      parameter (maxmat=10)
      common/csimpa/eabs(3,maxmat),c1(maxmat),c2(maxmat),wcc(maxmat),
     1  wcr(maxmat)
      real*8 p,st,dst,ds1,w1,w2,t1,t2
      common/cjump0/p(8),st,dst,ds1,w1,w2,t1,t2
      logical active,isLogScale
      integer maxbin,nbin,detmat
      parameter (maxbin=1000)
      real*8 flutmp,flu,flu2,egrid,ebingrd,emin,ebin,iebin,eratio
      real*8 unclimit
      common /scoftl/ flutmp(3,maxbin),flu(3,maxbin),flu2(3,maxbin),
     &                egrid(maxbin),ebingrd(maxbin),emin,ebin,iebin,
     &                eratio,unclimit,nbin,detmat,isLogScale,active
      logical called
      real*8 derest
      common /scofl0/ derest,called
      logical inside
      real*8 eini,ds,dssoft,desoft
      common /scofl1/ eini,ds,dssoft,desoft,inside
      integer i,k
      integer*4 ncross
      real*8 dseff

      if (.not.active) return

      if (mode.eq.3) then  ! STEP with no interface crossing, store JUMP distance for later use
        ds = arg
        return
      endif

      if (mode.lt.0.and.mode.gt.-10) then  ! Modes reserved for KNOCK
        ! Check KNOCKX usage:
        if (.not.called) then
          write(*,*)
     &      'FTLtally:INTERNAL ERROR: KNOCKX was not called.'
          stop
        endif

        ! A KNOCK has occurred, check if inside detection material:
        if (.not.inside) return

        if (mode.eq.-1.and.kpar.ne.2) then  ! Soft event, store partial jump distance and E-loss
          dssoft = ds
          desoft = derest

          if (e.lt.1.0) then  ! e+- was absorbed
            ! Correct soft energy loss assuming the e+- travelled just until E=Eabs:
            derest = eini-eabs(kpar,mat)
            ! Scale down the step length just until E=Eabs
            ! (variable DS1 contains the remaining substep length--from JUMP):
            ds1 = (dssoft+ds1)*derest/desoft-dssoft
            if (ds1.gt.0.0) then
              call step(ds1,dseff,ncross)
            else
              ! Scaled-down step shorter than 1st substep -> 2nd substep <0:
              dseff = ds1
            endif
            ! Tally step track length and restricted energy loss:
            ds = (dssoft+dseff)*wght
            call trackl(kpar,eini,derest,ds)
          endif
          return

        else  ! Tally 2nd part of a completed soft step--or a full step if a photon
          ds = (dssoft+ds)*wght
          call trackl(kpar,eini,desoft,ds)
          ! Reset accumulated soft vars after a hard collision:
          eini = e
          dssoft = 0.0
          desoft = 0.0
          return
        endif
      endif

      if (mode.eq.4) then ! Interface crossing, treated as a hard interaction
        if (inside) then  ! Tally step track length
          ds = (dssoft+arg)*wght
          call trackl(kpar,eini,desoft,ds)
          ! if we were inside the tally material, now we must be outside, so:
          inside = .false.
          return
        endif
        ! If we were not in the tally material, find out if we are now:
        if (mat.eq.detmat) then
          inside = .true.
          eini = e
          dssoft = 0.0
          desoft = 0.0
          return
        endif
      endif

      if (mode.eq.-99.or.mode.eq.0) then ! Stack modified: particle retrieved (-99) or put by SOURCE (0)
        if (mat.eq.detmat) then
          inside = .true.
          eini = e
          dssoft = 0.0
          desoft = 0.0
        else
          inside = .false.
        endif
        return
      endif

      if (mode.eq.6) then  ! End-of-History; dump and clear counters
        do i=1,nbin
          do k=1,3
            flu(k,i)    = flu(k,i) +flutmp(k,i)
            flu2(k,i)   = flu2(k,i)+flutmp(k,i)**2
            flutmp(k,i) = 0.0
          enddo
        enddo
        return
      endif

      if (mode.eq.-98.and.e.gt.eabs(kpar,mat).and.mat.eq.detmat)
     & then
        write(*,*)
     &    'FTLtally:ERROR: Particle absorbed inside the detector'
        write(*,*)
     &'             but its energy was higher than EABS; this'
        write(*,*)
     &'             prevents the calculation of the fluence down to'
        write(*,*)
     &'             EABS. Reformulate your particle rejection'
        write(*,*)
     &'             strategy appropriately.'
        stop
      endif
      end


      subroutine FTLreport(n,cputim,uncdone)
!*******************************************************************
!*    Input:                                                       *
!*      n -> no. of histories simulated                            *
!*      cputim -> elapsed CPU time                                 *
!*    Output:                                                      *
!*      uncdone -> 2 if uncert reached, 1 if not defined, 0 else   *
!*******************************************************************
      implicit none
      integer uncdone
      real*8 n,cputim

      integer*4 seed1,seed2
      common/rseed/seed1,seed2
      integer*4 kpar,ibody,mat,ilb
      real*8 e,x,y,z,u,v,w,wght
      common/track/e,x,y,z,u,v,w,wght,kpar,ibody,mat,ilb(5)
      logical active,isLogScale
      integer maxbin,nbin,detmat
      parameter (maxbin=1000)
      real*8 flutmp,flu,flu2,egrid,ebingrd,emin,ebin,iebin,eratio
      real*8 unclimit
      common /scoftl/ flutmp(3,maxbin),flu(3,maxbin),flu2(3,maxbin),
     &                egrid(maxbin),ebingrd(maxbin),emin,ebin,iebin,
     &                eratio,unclimit,nbin,detmat,isLogScale,active
      character*80 buffer
      integer i,nchan,out,finduf,error
      real*8 q,q2,sigma,eff,avesig,maxq,fact,emiddle,uncert,invn

      uncdone = 1
      if (.not.active) return

      call FTLdump(1)  ! Sim dump file

      do kpar=3,1,-1  ! One report for each particle type
        ! Prepare output files:
        out = finduf()
        if (kpar.eq.3)
     &  buffer = 'tallyFluenceTrackLength-positron.dat'
        if (kpar.eq.2)
     &  buffer = 'tallyFluenceTrackLength-photon.dat'
        if (kpar.eq.1)
     &  buffer = 'tallyFluenceTrackLength-electron.dat'
        open(out,file=buffer,iostat=error)
        if (error.ne.0) then
          write(*,*) ''
          write(*,'(a)')
     &      '*********************************************'
          write(*,'(a)')
     &      'FTLreport:ERROR: cannot open output data file'
          write(*,'(a)')
     &      '*********************************************'
          close(out)  ! Just in case
          return
        endif

        ! Evaluate 1/2 of the max fluence score:
        avesig = 0.0
        nchan = 0
        maxq = 0.0
        do i=1,nbin
          q = flu(kpar,i)/ebingrd(i)
          maxq = max(maxq,q)
        enddo
        maxq = 0.5*maxq

        write(out,'(a)')
     &'#>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>'
        write(out,'(a)') '# [SECTION REPORT FLUENCE SPECTRUM]'
        if (kpar.eq.3) buffer = '# POSITRONS'
        if (kpar.eq.2) buffer = '# PHOTONS'
        if (kpar.eq.1) buffer = '# ELECTRONS'
        write(out,'(a)') buffer
        write(out,'(a)') '#'
        write(out,'(a)')
     &  '# The differential fluence is integrated over the detector'
        write(out,'(a)')
     &  '#   volume, hence its units are cm^3/(cm^2*eV)=cm/eV'
        write(out,'(a)')
     &  '# For plotting purposes, two energies per bin are given,'
        write(out,'(a)')
     &  '#   namely, the low end and the middle point of each bin'
        write(out,'(a)') '#'
        write(out,'(a)')
     &  '# Elow(eV) : Emiddle(eV) : dFluence(cm/eV) : +-2sigma'
        invn = 1.0/n
        do i=1,nbin
          q = flu(kpar,i)*invn
          q2 = flu2(kpar,i)*invn
          sigma = (q2-q**2)*invn
          sigma = sqrt(max(sigma,0.0))
          fact = 1.0/ebingrd(i)
          q = q*fact
          sigma = sigma*fact
          if (isLogScale) then
            emiddle = egrid(i)*(eratio+1.0)*0.5  ! Middle energy in the bin
          else
            emiddle = egrid(i)+ebin*0.5
          endif
          write(out,'(3(1x,es12.5),1x,es8.1)')
     &      egrid(i),emiddle,q,2.0*sigma
          ! Evaluate average uncertainty for scores above 1/2 max score:
          if (flu(kpar,i)*fact.gt.maxq) then
            avesig = avesig+(sigma/q)**2
            nchan = nchan+1
          endif
        enddo
        uncdone = 0
        if (nchan.gt.0) then
          uncert = 200.0*sqrt(avesig/nchan)
          if (uncert.lt.unclimit) uncdone = 2  ! Uncertainty reached
        else
          uncert = 0.0  ! Uncertainty assumed not reached when score is nil
        endif

        ! Generic report:
        write(out,'(a)') ' '
        write(out,'(a)') '# Performance report'
        write(out,'(a)') '#   Random seeds:'
        write(out,'(a,i10)') '#   ',seed1
        write(out,'(a,i10)') '#   ',seed2
        write(out,'(a)') '#   No. of histories simulated [N]:'
        write(out,'(a,f18.0)') '#   ',n
        write(out,'(a)') '#   CPU time [t] (s):'
        write(out,'(a,es12.5)') '#   ',cputim
        if (cputim.gt.0.0) then
          write(out,'(a)') '#   Speed (histories/s):'
          write(out,'(a,es12.5)') '#   ',n/cputim
        endif
        write(out,'(a)')
     & '#   Average uncertainty (above 1/2 max score) in % [uncert]:'
        write(out,'(a,es12.5)') '#   ',uncert
        eff = n*uncert**2
        if (eff.gt.0.0) then
          write(out,'(a)') '#   Intrinsic efficiency [N*uncert^2]^-1:'
          write(out,'(a,es12.5)') '#   ',1.0/eff
        endif
        eff = cputim*uncert**2
        if (eff.gt.0.0) then
          write(out,'(a)') '#   Absolute efficiency [t*uncert^2]^-1:'
          write(out,'(a,es12.5)') '#   ',1.0/eff
        endif
        write(out,'(a)') '#'
        write(out,'(a)') '# Have a nice day.'
        close(out)
      enddo
      end


      subroutine FTLinitally
!*******************************************************************
!*    Initializes. To be called before TALLY.                      *
!*                                                                 *
!*    Comments:                                                    *
!*      -> Must be called *AFTER* PEINIT                           *
!*******************************************************************
      implicit none
      integer maxmat
      parameter (maxmat=10)
      real*8 eabs,c1,c2,wcc,wcr
      common/csimpa/eabs(3,maxmat),c1(maxmat),c2(maxmat),wcc(maxmat),
     1  wcr(maxmat)
      logical active,isLogScale
      integer maxbin,nbin,detmat
      parameter (maxbin=1000)
      real*8 flutmp,flu,flu2,egrid,ebingrd,emin,ebin,iebin,eratio
      real*8 unclimit
      common /scoftl/ flutmp(3,maxbin),flu(3,maxbin),flu2(3,maxbin),
     &                egrid(maxbin),ebingrd(maxbin),emin,ebin,iebin,
     &                eratio,unclimit,nbin,detmat,isLogScale,active
      logical called
      real*8 derest
      common /scofl0/ derest,called
      character*(*) secid,eos
      parameter (secid=
     &'[SECTION TALLY FLUENCE TRACK LENGTH v.2012-06-01]')
      parameter (eos='[END OF FTL SECTION]')
      character*80 buffer
      integer i,error
      real*8 emax

      write(*,*) ' '
      write(*,'(a)')
     & '>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>'
      call getline(buffer)
      if (index(buffer,secid).eq.0) then
        write(*,'(a)') 'FTLinitally:ERROR: incorrect section header;'
        write(*,'(a,a)') '  expecting to find: ',secid
        write(*,'(a,a)') '  found instead:     ',buffer
        stop
      endif
      write(*,'(a)') secid

      read(*,'(a80)') buffer
      buffer = adjustl(buffer)
      buffer = buffer(1:scan(buffer,' ')) ! Clip at 1st blank
      if (buffer.eq.'ON') then
        active = .true.
      else if (buffer.eq.'OFF') then
        active = .false.
        write(*, '(a)')
     &    '>>>> Tally Fluence Track Length is OFF >>>>'
        do
          read(*,'(a80)',iostat=error) buffer
          if (error.ne.0) then
            write(*,'(a,a,a)') 'FTLinitally:ERROR: ',
     &       'Unable to find End-Of-Section mark: ',eos
            stop
          endif
          if (index(buffer,eos).ne.0) return
        enddo
      else
        write(*,'(a)')
     &    'FTLinitally:ERROR: expecting to find ON or OFF'
        write(*,'(a)') 'found instead:'
        write(*,'(a)') buffer
        stop
      endif

      ! Read data:
      write(*,'(a)') 'Detection material set to:'
      read(*,*) detmat
      write(*,'(1x,i0)') detmat

      write(*,'(a)')
     &  'Spectrum limits [Emin,Emax] (eV) and no. bins:'
      read(*,'(a80)') buffer
      read(buffer,*) emin,emax,nbin,buffer
      buffer = adjustl(buffer)
      buffer = buffer(1:scan(buffer,' ')) ! Clip at 1st blank
      isLogScale = .false.
      if (buffer.eq.'LOG'.or.buffer.eq.'log') isLogScale = .true.
      write(*,'(2(1x,es12.5),1x,i0,$)') emin,emax,nbin
      nbin = nbin+1  ! One more bin to tally counts with E=Emax
      if (nbin.lt.2.or.nbin.gt.maxbin) then
        write(*,*) ''
        write(*,'(a)') 'FTLinitally:ERROR: Invalid no. of bins.'
        write(*,'(a,i0,1x,i0)') '  Min and max are ',1,maxbin-1
        stop
      endif

      if (isLogScale) then
        write(*,'(a)') ' log scale'
        eratio = (emax/emin)**(1.0d0/nbin)   ! Ratio of one grid point to the previous
        ebin = log(eratio)                 ! Bin width in log scale
        iebin = 1.0/ebin
        emin = log(emin)                   ! Store values is log scale
        do i=1,nbin                        ! Set energy grid
          egrid(i) = exp(emin+(i-1)*ebin)
          ebingrd(i) = egrid(i)*(eratio-1.0)
        enddo
      else
        write(*,'(a)') ' linear scale'
        eratio = 0.0                       ! Dummy value, not used, just in case
        ebin = (emax-emin)/nbin
        iebin = 1.0/ebin
        do i=1,nbin                        ! Set energy grid
          egrid(i) = emin+(i-1)*ebin
          ebingrd(i) = ebin
        enddo
      endif

      write(*,'(a)') 'Relative uncertainty (%) requested:'
      read(*,*) unclimit
      write(*,'(1x,es12.5)') unclimit

      ! Clear counters:
      flu    = 0.0
      flu2   = 0.0
      flutmp = 0.0

      ! Set KNOCKX test flag and reset soft Eloss:
      called = .false.
      derest = 0.0

      read(*,'(a80)') buffer
      if (index(buffer,eos).eq.0) then
        write(*,*) 'FTLinitally:ERROR: End-Of-Section mark not found'
        write(*,'(a,a)') '  expecting to find: ',eos
        write(*,'(a,a)') '  found instead:     ',buffer
        stop
      endif

      call FTLdump(0)  ! Sim restart file

      write(*,'(a)') '>>>> FTL tally initialization finished >>>>'
      end


      subroutine trackl(kpar,eini,de,ds)
!*******************************************************************
!*    Tallies track lengths, whence fluences are calculated.       *
!*    The energy grid can be logarithmic or linear.                *
!*                                                                 *
!*    Input:                                                       *
!*      kpar -> particle type                                      *
!*      eini -> initial particle energy (eV)                       *
!*      de -> energy loss along the step (eV)                      *
!*      ds -> track length (cm)                                    *
!*    Notes:                                                       *
!*      -> for an e+- DE and DS should refer to a hard-to-hard step*
!*******************************************************************
      implicit none
      integer*4 kpar
      real*8 eini,de,ds

      logical active,isLogScale
      integer maxbin,nbin,detmat
      parameter (maxbin=1000)
      real*8 flutmp,flu,flu2,egrid,ebingrd,emin,ebin,iebin,eratio
      real*8 unclimit
      common /scoftl/ flutmp(3,maxbin),flu(3,maxbin),flu2(3,maxbin),
     &                egrid(maxbin),ebingrd(maxbin),emin,ebin,iebin,
     &                eratio,unclimit,nbin,detmat,isLogScale,active
      integer i,binhi,binlo
      real*8 efin,istpav,dflu

      if (isLogScale) then  ! Energy grid in logarithmic scale

        binhi = (log(eini)-emin)*iebin+1.0
        if (de.gt.0.0) then
          ! Average inverse stopping power:
          istpav = ds/de
        else
          ! No need to spread over various bins:
          if (binhi.lt.1.or.binhi.gt.nbin) return
          flutmp(kpar,binhi) = flutmp(kpar,binhi)+ds
          return
        endif

        efin = eini-de
        binlo = (log(efin)-emin)*iebin+1.0
        ! Spread over all bins from Efin to Eini, skipping invalid bins:
        do i=max(binlo,1),min(binhi,nbin)
          flutmp(kpar,i) = flutmp(kpar,i)+ebingrd(i)*istpav
        enddo

        ! Subtract excess in limiting bins:
        if (binhi.gt.0.and.binhi.le.nbin)
     &    flutmp(kpar,binhi) = flutmp(kpar,binhi)-
     &                         (egrid(binhi)*eratio-eini)*istpav
        if (binlo.gt.0.and.binlo.le.nbin)
     &    flutmp(kpar,binlo) = flutmp(kpar,binlo)-
     &                         (efin-egrid(binlo))*istpav

      else  ! Energy grid is linear

        binhi = (eini-emin)*iebin+1.0
        if (de.gt.0.0) then
          ! Average inverse stopping power:
          istpav = ds/de
        else
          ! No need to spread over various bins:
          if (binhi.lt.1.or.binhi.gt.nbin) return
          flutmp(kpar,binhi) = flutmp(kpar,binhi)+ds
          return
        endif

        efin = eini-de
        binlo = (efin-emin)*iebin+1.0
        dflu = ebin*istpav
        ! Spread over all bins from Efin to Eini, skipping invalid bins:
        do i=max(binlo,1),min(binhi,nbin)
          flutmp(kpar,i) = flutmp(kpar,i)+dflu
        enddo

        ! Subtract excess in limiting bins:
        if (binhi.gt.0.and.binhi.le.nbin)
     &    flutmp(kpar,binhi) = flutmp(kpar,binhi)-
     &                         (egrid(binhi)+ebin-eini)*istpav
        if (binlo.gt.0.and.binlo.le.nbin)
     &    flutmp(kpar,binlo) = flutmp(kpar,binlo)-
     &                         (efin-egrid(binlo))*istpav

      endif
      end


      subroutine FTLdump(mode)
!*******************************************************************
!*    Dumps into or reads data from a dump file.                   *
!*                                                                 *
!*    Input:                                                       *
!*      mode -> 1 to write dump file, else to read from it.        *
!*******************************************************************
      implicit none
      integer mode

      logical active,isLogScale
      integer maxbin,nbin,detmat
      parameter (maxbin=1000)
      real*8 flutmp,flu,flu2,egrid,ebingrd,emin,ebin,iebin,eratio
      real*8 unclimit
      common /scoftl/ flutmp(3,maxbin),flu(3,maxbin),flu2(3,maxbin),
     &                egrid(maxbin),ebingrd(maxbin),emin,ebin,iebin,
     &                eratio,unclimit,nbin,detmat,isLogScale,active
      character*80 dumpfilen
      integer restartfile,dumpfile
      common /dump/ restartfile,dumpfile,dumpfilen

      if (mode.eq.1) then
        if (dumpfile.lt.0) return  ! No dump file open
        write(dumpfile) flu(1:3,1:nbin),flu2(1:3,1:nbin)
      else
        if (restartfile.lt.0) return  ! No restart file open
        read(restartfile) flu(1:3,1:nbin),flu2(1:3,1:nbin)
      endif
      end


!>>>> End Of File >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

