!*******************************************************************
!*                          TALLY                                  *
!*                    PHASE SPACE FILE                             *
!*                                                                 *
!* Short description:                                              *
!*   Writes to a file the dynamical state of all particles that    *
!*   reach a given material.                                       *
!*******************************************************************


      subroutine PSFtally(mode,arg)
!*******************************************************************
!*    Input:                                                       *
!*      mode -> Identifies the state of the calling procedure      *
!*      arg -> current history number (when mode=1)                *
!*******************************************************************
      implicit none
      integer mode
      real*8 arg

      integer*4 kpar,ibody,mat,ilb
      real*8 e,x,y,z,u,v,w,wght
      common/track/e,x,y,z,u,v,w,wght,kpar,ibody,mat,ilb(5)
      logical active
      integer detmat,psfunit,formatiaea
      real*8 nhist,nhlast,npar,nele,npos
      common /scopsf/ nhist,nhlast,npar,nele,npos,detmat,psfunit,
     &                formatiaea,active
      character*80 fmtstr
      parameter (fmtstr='(i0,8(1x,es12.5),6(1x,i0))')
      integer*4 source_write,dn
      common /scoiaeapsf/ source_write,dn

      if (.not.active) return

      if (mode.eq.4.or.mode.eq.-99) then  ! New material
        if (mat.ne.detmat) return
        dn = nhist-nhlast+0.5d0           ! 0.5 prevents truncation errors

        if (formatiaea.eq.0) then
          write(psfunit,fmtstr) kpar,e,x,y,z,u,v,w,wght,dn,
     &                        ilb(1),ilb(2),ilb(3),ilb(4),ilb(5)
        else
          ! call IAEAwrite  ! IAEA inactive
          continue
        endif

        nhlast = nhist
        npar = npar+1
        if (kpar.eq.1) then
          nele = nele+1
        else if (kpar.eq.3) then
          npos = npos+1
        endif
        ! mat = 0  ! Force main to stop this particle's simulation
        ! Particle is stopped by setting Eabs=+infty in the PSF detector

      else if (mode.eq.1.or.mode.eq.2) then  ! New history or hist. modified
        nhist = arg

      endif
      end


      subroutine PSFreport(n,cputim)
!*******************************************************************
!*    Input:                                                       *
!*      n -> no. of histories simulated                            *
!*      cputim -> elapsed CPU time                                 *
!*    Comments:                                                    *
!*      -> 'cputim' should not include initialization procedures;  *
!*         enter 0 or neg. if not available.                       *
!*******************************************************************
      implicit none
      real*8 n,cputim

      integer*4 seed1,seed2
      common/rseed/seed1,seed2
      logical active
      integer detmat,psfunit,formatiaea
      real*8 nhist,nhlast,npar,nele,npos
      common /scopsf/ nhist,nhlast,npar,nele,npos,detmat,psfunit,
     &                formatiaea,active
      integer out,finduf,error

      if (.not.active) return

      call PSFdump(1)  ! Sim dump file

      ! Prepare output files:
      out = finduf()
      open(out,file='tallyPhaseSpaceFile.dat',iostat=error)
      if (error.ne.0) then
        write(*,*)
        write(*,'(a)')
     &    '*********************************************'
        write(*,'(a)')
     &    'PSFreport:ERROR: cannot open output data file'
        write(*,'(a)')
     &    '*********************************************'
        close(out)  ! Just in case
        return
      endif

      write(out,'(a)')
     &'#>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>'
      write(out,'(a)') '# [SECTION REPORT PSF]'
      write(out,'(a)')
     &'# No. of electrons/photons/positrons/total written to PSF:'
      write(out,'(f18.0)') nele
      write(out,'(f18.0)') npar-nele-npos
      write(out,'(f18.0)') npos
      write(out,'(f18.0)') npar

      ! Generic report:
      write(out,*)
      write(out,'(a)') '# Performance report'
      write(out,'(a)') '#   Random seeds:'
      write(out,'(a,i10)') '#   ',seed1
      write(out,'(a,i10)') '#   ',seed2
      write(out,'(a)') '#   No. of histories simulated [N]:'
      write(out,'(a,f18.0)') '#   ',n
      write(out,'(a)') '#   CPU time [t] (s):'
      write(out,'(a,es12.5)') '#   ',cputim
      if (cputim.gt.0.0) then
        write(out,'(a)') '#   Speed (histories/s):'
        write(out,'(a,es12.5)') '#   ',n/cputim
      endif
      write(out,'(a)') '#'
      write(out,'(a)') '# Have a nice day.'
      close(out)
      end


      subroutine PSFinitally
!*******************************************************************
!*    Initializes. To be called before TALLY                       *
!*******************************************************************
      implicit none
      logical active
      integer detmat,psfunit,formatiaea
      real*8 nhist,nhlast,npar,nele,npos
      common /scopsf/ nhist,nhlast,npar,nele,npos,detmat,psfunit,
     &                formatiaea,active
      character*80 dumpfilen
      integer restartfile,dumpfile
      common /dump/ restartfile,dumpfile,dumpfilen
      character*(*) secid,eos
      parameter (secid=
     &'[SECTION TALLY PHASE SPACE FILE v.2009-06-15]')
      parameter (eos='[END OF PSF SECTION]')
      character*80 buffer
      integer finduf,error

      write(*,*) ' '
      write(*,'(a)')
     & '>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>'
      call getline(buffer)
      if (index(buffer,secid).eq.0) then
        write(*,'(a)') 'PSFinitally:ERROR: incorrect section header;'
        write(*,'(a,a)') '  expecting to find: ',secid
        write(*,'(a,a)') '  found instead:     ',buffer
        stop
      endif
      write(*,'(a)') secid

      read(*,'(a80)') buffer
      buffer = adjustl(buffer)
      buffer = buffer(1:scan(buffer,' ')) ! Clip at 1st blank
      if (buffer.eq.'ON') then
        active = .true.
      else if (buffer.eq.'OFF') then
        active = .false.
        write(*, '(a)')
     &    '>>>> Tally Phase Space File is OFF >>>>'
        do
          read(*,'(a80)',iostat=error) buffer
          if (error.ne.0) then
            write(*,'(a,a,a)') 'PSFinitally:ERROR: ',
     &       'Unable to find End-Of-Section mark: ',eos
            stop
          endif
          if (index(buffer,eos).ne.0) return
        enddo
      else
        write(*,'(a)')
     &    'PSFinitally:ERROR: expecting to find ON or OFF'
        write(*,'(a)') 'found instead:'
        write(*,'(a)') buffer
        stop
      endif

      read(*,*) formatiaea
      if (formatiaea.eq.0) then
        write(*,'(a)') 'PSF format: standard penEasy in ASCII.'
      else if (formatiaea.eq.1) then
        write(*,'(a)')
     &    'PSFinitally:ERROR: IAEA PSF format requested '//
     &    'but not available.'
        write(*,'(a)') '  Read the instructions provided in the '//
     &    'README file to activate this option.'
        stop
      else
        write(*,'(a)')
     &    'PSFinitally:ERROR: PSF format must be 0 or 1.'
        stop
      endif

      write(*,'(a)') 'Detection material set to:'
      read(*,*) detmat
      write(*,'(1x,i0)') detmat
      if (detmat.le.0) then
        write(*,*) 'PSFinitally:ERROR: detection material must be >0'
        stop
      endif

      write(*,'(a)') 'PSF filename:'
      read(*,'(a80)') buffer
      buffer = adjustl(buffer)
      buffer = buffer(1:scan(buffer,' ')) ! Clip at 1st blank
      write(*,'(1x,a)') buffer

      if (formatiaea.eq.0) then
        psfunit = finduf()
        write(*,'(a)') 'Opening PSF as unit:'
        write(*,'(1x,i0)') psfunit

        if (restartfile.lt.0) then  ! This is not a restart
          open(psfunit,file=buffer,status='replace')
          write(psfunit,'(a)')
     &      '# [PHASE SPACE FILE FORMAT penEasy v.2008-05-15]'
          write(psfunit,'(a)')
     &      '# KPAR : E : X : Y : Z : U : V : W : WGHT : '//
     &      'DeltaN : ILB(1..5)'
        else                        ! This is a restart, append to the PSF
          open(psfunit,file=buffer,status='old',position='append',
     &         iostat=error)
          if (error.ne.0) then
            write(*,'(a)') 'PSFinitally:ERROR: Could not append PSF;'//
     &        ' make sure the file exists and it is readable.'
            stop
          endif
          write(psfunit,'(a)')
     &      '# Restarting simulation, PSF continues...'
        endif
      else
        ! call IAEAwrite_ini(buffer)  ! IAEA inactive
        continue
      endif

      !*** Init vars:
      nhist  = 0.0d0
      nhlast = 0.0d0
      npar   = 0.0d0
      nele   = 0.0d0
      npos   = 0.0d0

      call PSFdump(0)   ! Sim restart file

      read(*,'(a80)') buffer
      if (index(buffer,eos).eq.0) then
        write(*,*) 'PSFinitally:ERROR: End-Of-Section mark not found'
        write(*,'(a,a)') '  expecting to find: ',eos
        write(*,'(a,a)') '  found instead:     ',buffer
        stop
      endif
      write(*,'(a)') '>>>> PSF tally initialization finished >>>>'
      end


      subroutine PSFdump(mode)
!*******************************************************************
!*    Dumps into or reads data from a dump file.                   *
!*                                                                 *
!*    Input:                                                       *
!*      mode -> 1 to write dump file, else to read from it.        *
!*******************************************************************
      implicit none
      integer mode

      logical active
      integer detmat,psfunit,formatiaea
      real*8 nhist,nhlast,npar,nele,npos
      common /scopsf/ nhist,nhlast,npar,nele,npos,detmat,psfunit,
     &                formatiaea,active
      character*80 dumpfilen
      integer restartfile,dumpfile
      common /dump/ restartfile,dumpfile,dumpfilen

      if (mode.eq.1) then
        if (dumpfile.lt.0) return  ! No dump file open
        write(dumpfile) nhlast,npar,nele,npos
      else
        if (restartfile.lt.0) return  ! No restart file open
        read(restartfile) nhlast,npar,nele,npos
      endif
      end


!>>>> End Of File >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
