!*******************************************************************
!*                           PENPATCH                              *
!*                                                                 *
!* Short description:                                              *
!*   Routines that supersede their standard PENELOPE counterparts. *
!*   The physics is NOT altered; the new routines only provide     *
!*   more information for some tallies to operate properly.        *
!*                                                                 *
!* Notes:                                                          *
!*   -> Parts that have been changed are written in lowercase and  *
!*      marked with 'cx'.                                          *
!*******************************************************************

!*******************************************************************
!* The block below is the PENELOPE copyright, which applies to the *
!* parts of the code extracted from the files included in the      *
!* official PENELOPE distribution.                                 *
!*******************************************************************
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C                                                                      C
C  PENELOPE/PENGEOM (version 2011)                                     C
C  Copyright (c) 2001-2011                                             C
C  Universitat de Barcelona                                            C
C                                                                      C
C  Permission to use, copy, modify, distribute and sell this software  C
C  and its documentation for any purpose is hereby granted without     C
C  fee, provided that the above copyright notice appears in all        C
C  copies and that both that copyright notice and this permission      C
C  notice appear in all supporting documentation. The Universitat de   C
C  Barelona makes no representations about the suitability of this     C
C  software for any purpose. It is provided "as is" without express    C
C  or implied warranty.                                                C
C                                                                      C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC


      subroutine JUMPx(dsmax,ds)
!*******************************************************************
!*    This routine provides a wrap-up for the standard JUMP and    *
!*    JUMPF of PENELOPE.                                           *
!*******************************************************************
      implicit none
      real*8 dsmax,ds

      integer*4 maxmat
      parameter (maxmat=10)
      logical analog,isforcing,active
      real*8 minwght
      common /frc001/ minwght,analog(maxmat,3),isforcing,active

      if (isforcing) then
        call jumpf(dsmax,ds)
      else
        call jump(dsmax,ds)
      endif
      end


      SUBROUTINE JUMPF(DSMAX,DS)
!*******************************************************************
!*    PENELOPE routine, no modifications introduced.               *
!*    Included here for convenience.                               *
!*******************************************************************
C
C  Modified subroutine 'JUMP' for interaction forcing.
C
C  Calculation of the free path from the starting point to the position
C  of the next event and of the probabilities of occurrence of different
C  events.
C
C  Arguments:
C    DSMAX ... maximum allowed step length (input),
C    DS ...... step length (output).
C
C  This subroutine does not modify the weight of the primary particle.
C
      IMPLICIT DOUBLE PRECISION (A-H,O-Z), INTEGER*4 (I-N)
C  ****  Main-PENELOPE common.
      COMMON/TRACK/E,X,Y,Z,U,V,W,WGHT,KPAR,IBODY,M,ILB(5)
C  ****  Simulation parameters.
      PARAMETER (MAXMAT=10)
      COMMON/CSIMPA/EABS(3,MAXMAT),C1(MAXMAT),C2(MAXMAT),WCC(MAXMAT),
     1  WCR(MAXMAT)
C  ****  Energy grid and interpolation constants for the current energy.
      PARAMETER (NEGP=200)
      COMMON/CEGRID/EMIN,EL,EU,ET(NEGP),DLEMP(NEGP),DLEMP1,DLFC,
     1  XEL,XE,XEK,KE
C  ****  Electron simulation tables.
      COMMON/CEIMFP/SEHEL(MAXMAT,NEGP),SEHIN(MAXMAT,NEGP),
     1  SEISI(MAXMAT,NEGP),SEHBR(MAXMAT,NEGP),SEAUX(MAXMAT,NEGP),
     2  SETOT(MAXMAT,NEGP),CSTPE(MAXMAT,NEGP),RSTPE(MAXMAT,NEGP),
     3  DEL(MAXMAT,NEGP),W1E(MAXMAT,NEGP),W2E(MAXMAT,NEGP),
     4  DW1EL(MAXMAT,NEGP),DW2EL(MAXMAT,NEGP),
     5  RNDCE(MAXMAT,NEGP),AE(MAXMAT,NEGP),BE(MAXMAT,NEGP),
     6  T1E(MAXMAT,NEGP),T2E(MAXMAT,NEGP)
C  ****  Positron simulation tables.
      COMMON/CPIMFP/SPHEL(MAXMAT,NEGP),SPHIN(MAXMAT,NEGP),
     1  SPISI(MAXMAT,NEGP),SPHBR(MAXMAT,NEGP),SPAN(MAXMAT,NEGP),
     2  SPAUX(MAXMAT,NEGP),SPTOT(MAXMAT,NEGP),CSTPP(MAXMAT,NEGP),
     3  RSTPP(MAXMAT,NEGP),W1P(MAXMAT,NEGP),W2P(MAXMAT,NEGP),
     4  DW1PL(MAXMAT,NEGP),DW2PL(MAXMAT,NEGP),
     5  RNDCP(MAXMAT,NEGP),AP(MAXMAT,NEGP),BP(MAXMAT,NEGP),
     6  T1P(MAXMAT,NEGP),T2P(MAXMAT,NEGP)
C  ****  Current IMFPs.
      COMMON/CJUMP0/P(8),ST,DST,DS1,W1,W2,T1,T2
      COMMON/CJUMP1/MODE,KSOFTE,KSOFTI,KDELTA
C  ****  Interaction forcing parameters.
      PARAMETER (NB=5000)
      COMMON/CFORCE/FORCE(NB,3,8)
      COMMON/CFORCG/P0(8),IFORC(8)
C
      EXTERNAL RAND
C
      IF(KPAR.EQ.1) THEN
C
C  ************  Electrons (KPAR=1).
C
        XEL=LOG(E)
        XE=1.0D0+(XEL-DLEMP1)*DLFC
        KE=XE
        XEK=XE-KE
        IF(MODE.EQ.1) THEN
          CALL EIMFP(1)
C  ****  Interaction forcing.
          DO KI=2,5
            IF(FORCE(IBODY,1,KI).GT.1.0D0.AND.P(KI).GT.1.0D-16) THEN
              P0(KI)=P(KI)
              P(KI)=P(KI)*FORCE(IBODY,1,KI)
              IFORC(KI)=1
            ELSE
              IFORC(KI)=0
            ENDIF
          ENDDO
          IFORC(1)=0
          IFORC(6)=0
          IFORC(7)=0
          IFORC(8)=0
          DS=DS1
          RETURN
        ENDIF
        CALL EIMFP(2)
C  ****  Interaction forcing.
        TFP=0.0D0
        DO KI=2,5
          IF(FORCE(IBODY,1,KI).GT.1.0D0.AND.P(KI).GT.1.0D-16) THEN
            P0(KI)=P(KI)
            P(KI)=P(KI)*FORCE(IBODY,1,KI)
            TFP=TFP+(P(KI)-P0(KI))
            IFORC(KI)=1
          ELSE
            IFORC(KI)=0
          ENDIF
        ENDDO
        IFORC(1)=0
        IFORC(6)=0
        IFORC(7)=0
        IFORC(8)=0
C
C  ****  Inverse hard mean free path (interaction probability per unit
C        path length).
C
        ST=P(2)+P(3)+P(4)+P(5)+P(8)
        DSMAXP=DSMAX
C
C  ****  Soft stopping interactions.
C        KSOFTI=1, soft stopping is active,
C        KSOFTI=0, soft stopping is not active.
C
        IF(W1.GT.1.0D-20) THEN
          KSOFTI=1
C  ****  The maximum step length, DSMAXP, is determined in terms of the
C        input DSMAX value (which is specified by the user) and the mean
C        free path for hard interactions (1/ST).
          DSMC=4.0D0/ST
          IF(DSMAXP.GT.DSMC) THEN
            DSMAXP=DSMC
          ELSE IF(DSMAXP.LT.1.0D-8) THEN
            DSMAXP=DSMC
          ENDIF
C  ****  The value of DSMAXP is randomized to eliminate dose artifacts
C        at the end of the first step.
          DSMAXP=(0.5D0+RAND(1.0D0)*0.5D0)*DSMAXP
C
C  ****  Upper bound for the interaction probability along the step
C        (including soft energy straggling).
C
          EDE0=W1*DSMAXP
          VDE0=W2*DSMAXP
          FSEDE=MAX(1.0D0-DW1EL(M,KE)*EDE0,0.75D0)
          FSVDE=MAX(1.0D0-DW2EL(M,KE)*EDE0,0.75D0)
          EDEM=EDE0*FSEDE
          VDEM=VDE0*FSVDE
          W21=VDEM/EDEM
          IF(EDEM.GT.9.0D0*W21) THEN
            ELOWER=MAX(E-(EDEM+3.0D0*SQRT(VDEM)),EMIN)
          ELSE IF(EDEM.GT.3.0D0*W21) THEN
            ELOWER=MAX(E-(EDEM+SQRT(3.0D0*VDEM)),EMIN)
          ELSE
            ELOWER=MAX(E-1.5D0*(EDEM+W21),EMIN)
          ENDIF
          XE1=1.0D0+(LOG(ELOWER)-DLEMP1)*DLFC
          KE1=XE1
          XEK1=XE1-KE1
          STLWR=EXP(SETOT(M,KE1)+(SETOT(M,KE1+1)-SETOT(M,KE1))*XEK1)
          ST=MAX(ST,STLWR+TFP)
        ELSE
          KSOFTI=0
        ENDIF
C
C  ****  Soft elastic scattering.
C        KSOFTE=1, soft scattering is active,
C        KSOFTE=0, soft scattering is not active.
C
        IF(T1.GT.1.0D-20) THEN
          KSOFTE=1
        ELSE
          KSOFTE=0
        ENDIF
C
C  ****  Delta interactions.
C        KDELTA=0, a hard interaction follows,
C        KDELTA=1, a delta interaction follows.
C
        DST=-LOG(RAND(3.0D0))/ST
        IF(DST.LT.DSMAXP) THEN
          KDELTA=0
        ELSE
          DST=DSMAXP
          KDELTA=1
        ENDIF
C
        IF(KSOFTE+KSOFTI.EQ.0) THEN
          DS=DST
          DS1=0.0D0
          MODE=1
        ELSE
          DS=DST*RAND(4.0D0)
          DS1=DST-DS
        ENDIF
        RETURN
      ELSE IF(KPAR.EQ.3) THEN
C
C  ************  Positrons (KPAR=3).
C
        XEL=LOG(E)
        XE=1.0D0+(XEL-DLEMP1)*DLFC
        KE=XE
        XEK=XE-KE
        IF(MODE.EQ.1) THEN
          CALL PIMFP(1)
C  ****  Interaction forcing.
          DO KI=2,6
            IF(FORCE(IBODY,3,KI).GT.1.0D0.AND.P(KI).GT.1.0D-16) THEN
              P0(KI)=P(KI)
              P(KI)=P(KI)*FORCE(IBODY,3,KI)
              IFORC(KI)=1
            ELSE
              IFORC(KI)=0
            ENDIF
          ENDDO
          IFORC(1)=0
          IFORC(7)=0
          IFORC(8)=0
          DS=DS1
          RETURN
        ENDIF
        CALL PIMFP(2)
C  ****  Interaction forcing.
        TFP=0.0D0
        DO KI=2,6
          IF(FORCE(IBODY,3,KI).GT.1.0D0.AND.P(KI).GT.1.0D-16) THEN
            P0(KI)=P(KI)
            P(KI)=P(KI)*FORCE(IBODY,3,KI)
            TFP=TFP+(P(KI)-P0(KI))
            IFORC(KI)=1
          ELSE
            IFORC(KI)=0
          ENDIF
        ENDDO
        IFORC(1)=0
        IFORC(7)=0
        IFORC(8)=0
C
C  ****  Inverse hard mean free path (interaction probability per unit
C        path length).
C
        ST=P(2)+P(3)+P(4)+P(5)+P(6)+P(8)
        DSMAXP=DSMAX
C
C  ****  Soft stopping interactions.
C        KSOFTI=1, soft stopping is active,
C        KSOFTI=0, soft stopping is not active.
C
        IF(W1.GT.1.0D-20) THEN
          KSOFTI=1
C  ****  The maximum step length, DSMAXP, is determined in terms of the
C        input DSMAX value (which is specified by the user) and the mean
C        free path for hard interactions (1/ST).
          DSMC=4.0D0/ST
          IF(DSMAXP.GT.DSMC) THEN
            DSMAXP=DSMC
          ELSE IF(DSMAXP.LT.1.0D-8) THEN
            DSMAXP=DSMC
          ENDIF
C  ****  The value of DSMAXP is randomized to eliminate dose artifacts
C        at the end of the first step.
          DSMAXP=(0.5D0+RAND(1.0D0)*0.5D0)*DSMAXP
C
C  ****  Upper bound for the interaction probability along the step
C        (including soft energy straggling).
C
          EDE0=W1*DSMAXP
          VDE0=W2*DSMAXP
          FSEDE=MAX(1.0D0-DW1PL(M,KE)*EDE0,0.75D0)
          FSVDE=MAX(1.0D0-DW2PL(M,KE)*EDE0,0.75D0)
          EDEM=EDE0*FSEDE
          VDEM=VDE0*FSVDE
          W21=VDEM/EDEM
          IF(EDEM.GT.9.0D0*W21) THEN
            ELOWER=MAX(E-(EDEM+3.0D0*SQRT(VDEM)),EMIN)
          ELSE IF(EDEM.GT.3.0D0*W21) THEN
            ELOWER=MAX(E-(EDEM+SQRT(3.0D0*VDEM)),EMIN)
          ELSE
            ELOWER=MAX(E-1.5D0*(EDEM+W21),EMIN)
          ENDIF
          XE1=1.0D0+(LOG(ELOWER)-DLEMP1)*DLFC
          KE1=XE1
          XEK1=XE1-KE1
          STLWR=EXP(SPTOT(M,KE1)+(SPTOT(M,KE1+1)-SPTOT(M,KE1))*XEK1)
          ST=MAX(ST,STLWR+TFP)
        ELSE
          KSOFTI=0
        ENDIF
C
C  ****  Soft elastic scattering.
C        KSOFTE=1, soft scattering is active,
C        KSOFTE=0, soft scattering is not active.
C
        IF(T1.GT.1.0D-20) THEN
          KSOFTE=1
        ELSE
          KSOFTE=0
        ENDIF
C
C  ****  Delta interactions.
C        KDELTA=0, a hard interaction follows,
C        KDELTA=1, a delta interaction follows.
C
        DST=-LOG(RAND(3.0D0))/ST
        IF(DST.LT.DSMAXP) THEN
          KDELTA=0
        ELSE
          DST=DSMAXP
          KDELTA=1
        ENDIF
C
        IF(KSOFTE+KSOFTI.EQ.0) THEN
          DS=DST
          DS1=0.0D0
          MODE=1
        ELSE
          DS=DST*RAND(4.0D0)
          DS1=DST-DS
        ENDIF
        RETURN
      ELSE
C
C  ************  Photons (KPAR=2).
C
        IF(MODE.EQ.1) THEN  ! Energy is the same as in previous calls.
          DS=-DLOG(RAND(1.0D0))/ST  ! IMFPs already computed.
          RETURN
        ENDIF
C
        MODE=0  ! New energy. Full calculation.
        XEL=LOG(E)
        XE=1.0D0+(XEL-DLEMP1)*DLFC
        KE=XE
        XEK=XE-KE
C
        CALL GIMFP
C  ****  Interaction forcing.
        DO KI=1,4
          IF(FORCE(IBODY,2,KI).GT.1.0D0.AND.P(KI).GT.1.0D-16) THEN
            P0(KI)=P(KI)
            P(KI)=P(KI)*FORCE(IBODY,2,KI)
            IFORC(KI)=1
          ELSE
            IFORC(KI)=0
          ENDIF
        ENDDO
        IFORC(5)=0
        IFORC(6)=0
        IFORC(7)=0
        IFORC(8)=0
C
        ST=P(1)+P(2)+P(3)+P(4)+P(8)
        DS=-LOG(RAND(1.0D0))/ST
      ENDIF
      RETURN
      END


      SUBROUTINE KNOCKx(DEF,ICOL)
!*******************************************************************
!*   This routine supersedes the standard KNOCKF of PENELOPE and it*
!*   also provides an overall wrap-up for both KNOCK and KNOCKF,   *
!*   invoking one or the other depending on the status of the      *
!*   interaction forcing switch.                                   }
!*******************************************************************
C
C  Modified subroutine 'KNOCK' for interaction forcing.
C
C  Output arguments:
C    DEF .... effective energy deposited by the particle in the mater-
C             ial. Includes a weight correction such that no special
C             action needs to be taken in the main program.
C             --> Use with care; when this quantity is considered as
C             the deposited energy, simulated energy deposition spectra
C             are biased.
C    ICOL ... kind of interaction experienced by the particle.
C
C  This subroutine does not modify the weight of the primary particle.
C
      IMPLICIT DOUBLE PRECISION (A-H,O-Z), INTEGER*4 (I-N)
C  ****  Main-PENELOPE commons.
      COMMON/TRACK/E,X,Y,Z,U,V,W,WGHT,KPAR,IBODY,M,ILB(5)
      COMMON/STOKES/SP1,SP2,SP3,IPOL
C  ****  Simulation parameters.
      PARAMETER (MAXMAT=10)
      COMMON/CSIMPA/EABS(3,MAXMAT),C1(MAXMAT),C2(MAXMAT),WCC(MAXMAT),
     1  WCR(MAXMAT)
C  ****  Secondary stack.
      PARAMETER (NMS=1000)
      COMMON/SECST/ES(NMS),XS(NMS),YS(NMS),ZS(NMS),US(NMS),
     1   VS(NMS),WS(NMS),WGHTS(NMS),SP1S(NMS),SP2S(NMS),SP3S(NMS),
     2   KS(NMS),IBODYS(NMS),MS(NMS),ILBS(5,NMS),IPOLS(NMS),NSEC
      COMMON/CERSEC/IERSEC
C  ****  Current IMFPs.
      COMMON/CJUMP0/P(8),ST,DST,DS1,W1,W2,T1,T2
      COMMON/CJUMP1/MODE,KSOFTE,KSOFTI,KDELTA
C  ****  Interaction forcing parameters.
      COMMON/CFORCG/P0(8),IFORC(8)
C
      EXTERNAL RAND

!cx:   *** This common is used to extract info on the restricted Eloss:
      logical called
      real*8 derest
      common /scofl0/ derest,called

!cx:   *** This common contains information about interaction forcing:
      logical analog,isforcing,active
      real*8 minwght
      common /frc001/ minwght,analog(maxmat,3),isforcing,active

!cx: Check interaction forcing:
      if (.not.isforcing) then
        call knocki(def,icol)
        return
      endif

C
C  ****  Store particle state variables before the interaction.
C
      EA=E
      UA=U
      VA=V
      WA=W
      SP1A=SP1
      SP2A=SP2
      SP3A=SP3
      IPOLA=IPOL
      NSECA=NSEC
C
C  ****  Simulate a new interaction.
C
!cx:  CALL KNOCK(DE,ICOL)
      call knocki(de,icol)
      IF(IFORC(ICOL).EQ.0) THEN  ! Unforced interaction.
        DEF=DE
        RETURN
      ENDIF
C  ****  Modify the weights of generated secondary particles, if any.
      WFORCE=P0(ICOL)/P(ICOL)
      DEF=DE*WFORCE

!cx:
      derest = derest*wforce

      IF(NSEC.GT.NSECA) THEN
C  ****  Allow writing multiple stack overflow warnings.
        IF(IERSEC.NE.0) IERSEC=0
        DO I=NSECA+1,NSEC
          WGHTS(I)=WGHTS(I)*WFORCE
        ENDDO
      ENDIF
C  ****  And set the final state of the primary particle.
      IF(RAND(1.0D0).GT.WFORCE) THEN
C  ****  The primary particle state is not affected.
        E=EA
        U=UA
        V=VA
        W=WA
        SP1=SP1A
        SP2=SP2A
        SP3=SP3A
        IPOL=IPOLA
        IF(KPAR.EQ.2) MODE=1  ! Same energy as in previous calls.
      ELSE
        IF(KPAR.EQ.2) THEN
          IF(ABS(E-EA).LT.1.0D0) THEN
            MODE=1  ! Same energy as in previous calls.
          ELSE
            MODE=0
          ENDIF
        ENDIF
      ENDIF
C
      RETURN
      END


      SUBROUTINE KNOCKi(DE,ICOL)
!*******************************************************************
!*   This routine supersedes the standard KNOCK of PENELOPE.       *
!*   Its code has been altered only to extract information needed  *
!*   to tally the fluence using the track length estimator.        *
!*   This information is accessed via a common block.              *
!*   The physics of PENELOPE has not been modified in any way.     *
!*******************************************************************
C
C     Simulation of random hinges and hard interaction events.
C
C  Output arguments:
C    DE ..... energy deposited by the particle in the material. It is
C             usually equal to the difference between the energies
C             before and after the interaction.
C    ICOL ... kind of interaction suffered by the particle.
C
      IMPLICIT DOUBLE PRECISION (A-H,O-Z), INTEGER*4 (I-N)
      CHARACTER*2 LASYMB
      PARAMETER (PI=3.1415926535897932D0, TWOPI=PI+PI)
      PARAMETER (REV=5.10998918D5)  ! Electron rest energy (eV)
      PARAMETER (RREV=1.0D0/REV, TREV=2.0D0*REV)
      PARAMETER (TRUNC=1.01538698D0)
C  ****  Main-PENELOPE commons.
      COMMON/TRACK/E,X,Y,Z,U,V,W,WGHT,KPAR,IBODY,M,ILB(5)
      COMMON/STOKES/SP1,SP2,SP3,IPOL
      COMMON/CHIST/ILBA(5)
C  ****  Simulation parameters.
      PARAMETER (MAXMAT=10)
      COMMON/CSIMPA/EABS(3,MAXMAT),C1(MAXMAT),C2(MAXMAT),WCC(MAXMAT),
     1  WCR(MAXMAT)
C  ****  Composition data.
      COMMON/COMPOS/STF(MAXMAT,30),ZT(MAXMAT),AT(MAXMAT),RHO(MAXMAT),
     1  VMOL(MAXMAT),IZ(MAXMAT,30),NELEM(MAXMAT)
C  ****  Energy grid and interpolation constants for the current energy.
      PARAMETER (NEGP=200)
      COMMON/CEGRID/EMIN,EL,EU,ET(NEGP),DLEMP(NEGP),DLEMP1,DLFC,
     1  XEL,XE,XEK,KE
C  ****  Element data.
      COMMON/CADATA/ATW(99),EPX(99),RSCR(99),ETA(99),EB(99,30),
     1  IFI(99,30),IKS(99,30),NSHT(99),LASYMB(99)
C  ****  E/P inelastic collisions.
      PARAMETER (NO=128)
      COMMON/CEIN/EXPOT(MAXMAT),OP2(MAXMAT),F(MAXMAT,NO),UI(MAXMAT,NO),
     1  WRI(MAXMAT,NO),KZ(MAXMAT,NO),KS(MAXMAT,NO),NOSC(MAXMAT)
C  ****  Compton scattering.
      PARAMETER (NOCO=128)
      COMMON/CGCO/FCO(MAXMAT,NOCO),UICO(MAXMAT,NOCO),FJ0(MAXMAT,NOCO),
     2  KZCO(MAXMAT,NOCO),KSCO(MAXMAT,NOCO),NOSCCO(MAXMAT)
C  ****  Bremsstrahlung emission.
      PARAMETER (NBW=32)
      COMMON/CEBR/WB(NBW),PBCUT(MAXMAT,NEGP),WBCUT(MAXMAT,NEGP),
     1  PDFB(MAXMAT,NEGP,NBW),DPDFB(MAXMAT,NEGP,NBW),
     2  PACB(MAXMAT,NEGP,NBW),ZBR2(MAXMAT)
C  ****  Electron simulation tables.
      COMMON/CEIMFP/SEHEL(MAXMAT,NEGP),SEHIN(MAXMAT,NEGP),
     1  SEISI(MAXMAT,NEGP),SEHBR(MAXMAT,NEGP),SEAUX(MAXMAT,NEGP),
     2  SETOT(MAXMAT,NEGP),CSTPE(MAXMAT,NEGP),RSTPE(MAXMAT,NEGP),
     3  DEL(MAXMAT,NEGP),W1E(MAXMAT,NEGP),W2E(MAXMAT,NEGP),
     4  DW1EL(MAXMAT,NEGP),DW2EL(MAXMAT,NEGP),
     5  RNDCE(MAXMAT,NEGP),AE(MAXMAT,NEGP),BE(MAXMAT,NEGP),
     6  T1E(MAXMAT,NEGP),T2E(MAXMAT,NEGP)
C  ****  Positron simulation tables.
      COMMON/CPIMFP/SPHEL(MAXMAT,NEGP),SPHIN(MAXMAT,NEGP),
     1  SPISI(MAXMAT,NEGP),SPHBR(MAXMAT,NEGP),SPAN(MAXMAT,NEGP),
     2  SPAUX(MAXMAT,NEGP),SPTOT(MAXMAT,NEGP),CSTPP(MAXMAT,NEGP),
     3  RSTPP(MAXMAT,NEGP),W1P(MAXMAT,NEGP),W2P(MAXMAT,NEGP),
     4  DW1PL(MAXMAT,NEGP),DW2PL(MAXMAT,NEGP),
     5  RNDCP(MAXMAT,NEGP),AP(MAXMAT,NEGP),BP(MAXMAT,NEGP),
     6  T1P(MAXMAT,NEGP),T2P(MAXMAT,NEGP)
C  ****  Current IMFPs.
      COMMON/CJUMP0/P(8),ST,DST,DS1,W1,W2,T1,T2
      COMMON/CJUMP1/MODE,KSOFTE,KSOFTI,KDELTA
C
      COMMON/CELSEP/EELMAX(MAXMAT),PELMAX(MAXMAT),
     1              RNDCEd(MAXMAT,NEGP),RNDCPd(MAXMAT,NEGP)
C
      EXTERNAL RAND
C
!cx:   *** This common us used to extract info on the restricted Eloss:
      logical called
      real*8 derest
      common /scofl0/ derest,called

!cx: (default)
      derest = 0.0
!cx: This is to test that KNOCKi has been used:
      called = .true.

      IF(KPAR.EQ.2) GO TO 2000
      IF(KPAR.EQ.3) GO TO 3000
C
C  ************  Electrons (KPAR=1).
C
C1000 CONTINUE
      IF(MODE.EQ.1) GO TO 1100
C
C  ****  Artificial soft event (ICOL=1).
C
      ICOL=1
      MODE=1
C
      IF(KSOFTI.EQ.0) THEN
        DE=0.0D0
      ELSE
        EDE0=W1*DST
        VDE0=W2*DST
        FSEDE=MAX(1.0D0-DW1EL(M,KE)*EDE0,0.75D0)
        FSVDE=MAX(1.0D0-DW2EL(M,KE)*EDE0,0.75D0)
        EDE=EDE0*FSEDE
        VDE=VDE0*FSVDE
C  ****  Generation of random values DE with mean EDE and variance VDE.
        SIGMA=SQRT(VDE)
        IF(SIGMA.LT.0.333333333D0*EDE) THEN
C  ****  Truncated Gaussian distribution.
          DE=EDE+RNDG3()*SIGMA
        ELSE
          RU=RAND(1.0D0)
          EDE2=EDE*EDE
          VDE3=3.0D0*VDE
          IF(EDE2.LT.VDE3) THEN
            PNULL=(VDE3-EDE2)/(VDE3+3.0D0*EDE2)
            IF(RU.LT.PNULL) THEN
              DE=0.0D0
            ELSE
              DE=1.5D0*(EDE+VDE/EDE)*(RU-PNULL)/(1.0D0-PNULL)
            ENDIF
          ELSE
            DE=EDE+(2.0D0*RU-1.0D0)*SQRT(VDE3)
          ENDIF
        ENDIF
      ENDIF
C
      E=E-DE
!cx:
      derest = de
      IF(E.LT.EABS(1,M)) THEN
        DE=E+DE
        E=0.0D0
        RETURN
      ENDIF
      IF(KSOFTE.EQ.0) RETURN
C
C  ****  Bielajew's randomly alternate hinge.
C
      IF(DE.GT.1.0D-3) THEN
        IF(RAND(2.0D0)*DST.GT.DS1) THEN
          XEL=LOG(E)
          XE=1.0D0+(XEL-DLEMP1)*DLFC
          KE=XE
          XEK=XE-KE
          IF(T1E(M,KE+1).GT.-78.3D0) THEN
            T1=EXP(T1E(M,KE)+(T1E(M,KE+1)-T1E(M,KE))*XEK)
            T2=EXP(T2E(M,KE)+(T2E(M,KE+1)-T2E(M,KE))*XEK)
          ELSE
            T1=0.0D0
            T2=0.0D0
          ENDIF
          IF(T1.LT.1.0D-35) RETURN
        ENDIF
      ENDIF
C  ****  1st and 2nd moments of the angular distribution.
      EMU1=0.5D0*(1.0D0-EXP(-DST*T1))
      EMU2=EMU1-(1.0D0-EXP(-DST*T2))/6.0D0
C  ****  Sampling from a two-bar histogram with these moments.
      PNUM=2.0D0*EMU1-3.0D0*EMU2
      PDEN=1.0D0-2.0D0*EMU1
      PB=PNUM/PDEN
      PA=PDEN+PB
      RND=RAND(3.0D0)
      IF(RND.LT.PA) THEN
        CDT=1.0D0-2.0D0*PB*(RND/PA)
      ELSE
        CDT=1.0D0-2.0D0*(PB+(1.0D0-PB)*((RND-PA)/(1.0D0-PA)))
      ENDIF
      DF=TWOPI*RAND(4.0D0)
      CALL DIRECT(CDT,DF,U,V,W)
      RETURN
C
C  ************  Hard event.
C
 1100 CONTINUE
      MODE=0
C  ****  A delta interaction (ICOL=7) occurs when the maximum
C        allowed step length is exceeded.
      IF(KDELTA.EQ.1) THEN
        ICOL=7
        DE=0.0D0
        RETURN
      ENDIF
C  ****  Interaction probabilities.
      STNOW=P(2)+P(3)+P(4)+P(5)
C  ****  Random sampling of the interaction type.
      STS=MAX(STNOW,ST)*RAND(5.0D0)
      SS=P(2)
      IF(SS.GT.STS) GO TO 1200
      SS=SS+P(3)
      IF(SS.GT.STS) GO TO 1300
      SS=SS+P(4)
      IF(SS.GT.STS) GO TO 1400
      SS=SS+P(5)
      IF(SS.GT.STS) GO TO 1500
      SS=SS+P(8)
      IF(SS.GT.STS) GO TO 1800
C  ****  A delta interaction (ICOL=7) may occur when the total
C        interaction probability per unit path length, ST, is
C        larger than STNOW.
      ICOL=7
      DE=0.0D0
      RETURN
C
C  ****  Hard elastic collision (ICOL=2).
C
 1200 ICOL=2
      IF(E.GE.EELMAX(M)) THEN
        TRNDC=RNDCE(M,KE)+(RNDCE(M,KE+1)-RNDCE(M,KE))*XEK
        TA=EXP(AE(M,KE)+(AE(M,KE+1)-AE(M,KE))*XEK)
        TB=BE(M,KE)+(BE(M,KE+1)-BE(M,KE))*XEK
        CALL EELa(TA,TB,TRNDC,RMU)
      ELSE
        TRNDC=RNDCEd(M,KE)+(RNDCEd(M,KE+1)-RNDCEd(M,KE))*XEK
        CALL EELd(TRNDC,RMU)  ! Uses the ELSEPA database.
      ENDIF
      CDT=1.0D0-(RMU+RMU)
      DF=TWOPI*RAND(6.0D0)
      CALL DIRECT(CDT,DF,U,V,W)
      DE=0.0D0
      RETURN
C
C  ****  Hard inelastic collision (ICOL=3).
C
 1300 ICOL=3
      DELTA=DEL(M,KE)+(DEL(M,KE+1)-DEL(M,KE))*XEK
      CALL EINa(E,DELTA,DE,EP,CDT,ES,CDTS,M,IOSC)
!cx:
      if (de.lt.eabs(1,m)) derest = de
C  ****  Scattering angles (primary electron).
      DF=TWOPI*RAND(7.0D0)
C  ****  Delta ray.
      IF(ES.GT.EABS(1,M)) THEN
        DFS=DF+PI
        US=U
        VS=V
        WS=W
        CALL DIRECT(CDTS,DFS,US,VS,WS)
        ILBA(1)=ILB(1)+1
        ILBA(2)=KPAR
        ILBA(3)=ICOL
        ILBA(4)=0
        ILBA(5)=ILB(5)
        CALL STORES(ES,X,Y,Z,US,VS,WS,WGHT,1,ILBA,0)
      ENDIF
C  ****  New energy and direction.
      IF(EP.GT.EABS(1,M)) THEN
        E=EP
        CALL DIRECT(CDT,DF,U,V,W)
        RETURN
      ENDIF
      DE=E
      E=0.0D0
      RETURN
C
C  ****  Hard bremsstrahlung emission (ICOL=4).
C
 1400 ICOL=4
      CALL EBRa(E,DE,M)
C  ****  Bremsstrahlung photon.
      IF(DE.GT.EABS(2,M)) THEN
        CALL EBRaA(E,DE,CDTS,M)
        DFS=TWOPI*RAND(8.0D0)
        US=U
        VS=V
        WS=W
        CALL DIRECT(CDTS,DFS,US,VS,WS)
        ILBA(1)=ILB(1)+1
        ILBA(2)=KPAR
        ILBA(3)=ICOL
        ILBA(4)=0
        ILBA(5)=ILB(5)
        CALL STORES(DE,X,Y,Z,US,VS,WS,WGHT,2,ILBA,0)
!cx:
      else
        derest = de
      ENDIF
C  ****  New energy.
      E=E-DE
      IF(E.GT.EABS(1,M)) RETURN
      DE=E+DE
      E=0.0D0
      RETURN
C
C  ****  Ionisation of an inner shell (ICOL=5) -does not affect the
C        primary electron.
C
 1500 ICOL=5
      DE=0.0D0
      CALL ESIa(IZA,ISA)
      IF(IZA.LT.1) RETURN
      ILBA(3)=ICOL
      CALL RELAX(IZA,ISA)
      RETURN
C
C  ****  Auxiliary fictitious mechanism (ICOL=8).
C
 1800 ICOL=8
      DE=0.0D0
      CALL EAUX
      RETURN
C
C  ************  Photons (KPAR=2).
C
 2000 CONTINUE
C
      STS=ST*RAND(1.0D0)
      SS=P(1)
      IF(SS.GT.STS) GO TO 2100
      SS=SS+P(2)
      IF(SS.GT.STS) GO TO 2200
      SS=SS+P(3)
      IF(SS.GT.STS) GO TO 2300
      SS=SS+P(4)
      IF(SS.GT.STS) GO TO 2400
      SS=SS+P(8)
      IF(SS.GT.STS) GO TO 2800
C
C  ****  Rayleigh scattering (ICOL=1).
C
 2100 CONTINUE
      DE=0.0D0
      CALL GRAa(E,CDT,M,IEFF)
C  ****  Delta interaction. Introduced to correct for the use of an
C        upper bound of the Rayleigh attenuation coefficient.
      IF(IEFF.EQ.0) THEN
        ICOL=7
        RETURN
      ENDIF
      ICOL=1
      IF(IPOL.EQ.1) THEN
        CALL DIRPOL(CDT,DF,0.0D0,SP1,SP2,SP3,U,V,W)
      ELSE
        DF=TWOPI*RAND(2.0D0)
        CALL DIRECT(CDT,DF,U,V,W)
      ENDIF
      RETURN
C
C  ****  Compton scattering (ICOL=2).
C
 2200 ICOL=2
      CALL GCOa(E,DE,EP,CDT,ES,CDTS,M,ISHELL)
      IZA=KZCO(M,ISHELL)
      ISA=KSCO(M,ISHELL)
      US=U
      VS=V
      WS=W
      DF=-1.0D0
      IF(IZA.GT.0.AND.ISA.LT.10) THEN
        ILBA(3)=ICOL
        CALL RELAX(IZA,ISA)
      ENDIF
C  ****  New direction and energy.
      IF(EP.GT.EABS(2,M)) THEN
        IF(IPOL.EQ.1) THEN
          ECDT=E*RREV*(1.0D0-CDT)
          CONS=ECDT*ECDT/(1.0D0+ECDT)
          CALL DIRPOL(CDT,DF,CONS,SP1,SP2,SP3,U,V,W)
        ELSE
          DF=TWOPI*RAND(3.0D0)
          CALL DIRECT(CDT,DF,U,V,W)
        ENDIF
        E=EP
      ELSE
        DE=E
        E=0.0D0
      ENDIF
C  ****  Compton electron.
      IF(ES.GT.EABS(1,M)) THEN
        IF(DF.LT.-0.5D0) DF=TWOPI*RAND(4.0D0)
        DFS=DF+PI
        CALL DIRECT(CDTS,DFS,US,VS,WS)
        ILBA(1)=ILB(1)+1
        ILBA(2)=KPAR
        ILBA(3)=ICOL
        ILBA(4)=IZA*1000000+ISA
        ILBA(5)=ILB(5)
        CALL STORES(ES,X,Y,Z,US,VS,WS,WGHT,1,ILBA,0)
      ENDIF
      RETURN
C
C  ****  Photoelectric absorption (ICOL=3).
C
 2300 ICOL=3
      CALL GPHa(ES,IZA,ISA)
C  ****  Delta interaction. Introduced to correct for the use of an
C        upper bound of the photoelectric attenuation coefficient.
      IF(IZA.EQ.0) THEN
        ICOL=7
        DE=0.0D0
        RETURN
      ENDIF
C
      IF(ES.GT.EABS(1,M)) THEN
        CALL SAUTER(ES,CDTS)
        DFS=TWOPI*RAND(5.0D0)
        US=U
        VS=V
        WS=W
        CALL DIRECT(CDTS,DFS,US,VS,WS)
        ILBA(1)=ILB(1)+1
        ILBA(2)=KPAR
        ILBA(3)=ICOL
        ILBA(4)=IZA*1000000+ISA
        ILBA(5)=ILB(5)
        CALL STORES(ES,X,Y,Z,US,VS,WS,WGHT,1,ILBA,0)
      ENDIF
      IF(ISA.LT.17) THEN
        ILBA(3)=ICOL
        CALL RELAX(IZA,ISA)
      ENDIF
      DE=E
      E=0.0D0
      RETURN
C
C  ****  Electron-positron pair production (ICOL=4).
C
 2400 ICOL=4
      CALL GPPa(EE,CDTE,EP,CDTP,IZA,ISA)
      DE=E
C  ****  Electron.
      IF(EE.GT.EABS(1,M)) THEN
        DF=TWOPI*RAND(6.0D0)
        US=U
        VS=V
        WS=W
        CALL DIRECT(CDTE,DF,US,VS,WS)
        ILBA(1)=ILB(1)+1
        ILBA(2)=KPAR
        ILBA(3)=ICOL
        ILBA(4)=0
        ILBA(5)=ILB(5)
        CALL STORES(EE,X,Y,Z,US,VS,WS,WGHT,1,ILBA,0)
      ENDIF
C  ****  Positron.
      IF(EP.GT.EABS(3,M)) THEN
        DF=TWOPI*RAND(7.0D0)
        CALL DIRECT(CDTP,DF,U,V,W)
        ILBA(1)=ILB(1)+1
        ILBA(2)=KPAR
        ILBA(3)=ICOL
        ILBA(4)=0
        ILBA(5)=ILB(5)
        CALL STORES(EP,X,Y,Z,U,V,W,WGHT,3,ILBA,0)
C  ****  The positron carries a 'latent' energy of 1022 keV.
        DE=DE-TREV
      ELSE
        CALL PANaR
      ENDIF
      E=0.0D0
C  ****  Atomic relaxation after triplet production.
      IF(ISA.LT.17) THEN
        ILBA(3)=ICOL
        CALL RELAX(IZA,ISA)
      ENDIF
      RETURN
C
C  ****  Auxiliary fictitious mechanism (ICOL=8).
C
 2800 ICOL=8
      DE=0.0D0
      CALL GAUX
      RETURN
C
C  ************  Positrons (KPAR=3).
C
 3000 CONTINUE
      IF(MODE.EQ.1) GO TO 3100
C
C  ****  Artificial soft event (ICOL=1).
C
      ICOL=1
      MODE=1
C
      IF(KSOFTI.EQ.0) THEN
        DE=0.0D0
      ELSE
        EDE0=W1*DST
        VDE0=W2*DST
        FSEDE=MAX(1.0D0-DW1PL(M,KE)*EDE0,0.75D0)
        FSVDE=MAX(1.0D0-DW2PL(M,KE)*EDE0,0.75D0)
        EDE=EDE0*FSEDE
        VDE=VDE0*FSVDE
C  ****  Generation of random values DE with mean EDE and variance VDE.
        SIGMA=SQRT(VDE)
        IF(SIGMA.LT.0.333333333D0*EDE) THEN
C  ****  Truncated Gaussian distribution.
          DE=EDE+RNDG3()*SIGMA
        ELSE
          RU=RAND(1.0D0)
          EDE2=EDE*EDE
          VDE3=3.0D0*VDE
          IF(EDE2.LT.VDE3) THEN
            PNULL=(VDE3-EDE2)/(VDE3+3.0D0*EDE2)
            IF(RU.LT.PNULL) THEN
              DE=0.0D0
            ELSE
              DE=1.5D0*(EDE+VDE/EDE)*(RU-PNULL)/(1.0D0-PNULL)
            ENDIF
          ELSE
            DE=EDE+(2.0D0*RU-1.0D0)*SQRT(VDE3)
          ENDIF
        ENDIF
      ENDIF
C
      E=E-DE
!cx:
      derest = de
C  ****  Annihilation at rest.
      IF(E.LT.EABS(3,M)) THEN
        DE=E+DE+TREV
        E=0.0D0
        CALL PANaR
        RETURN
      ENDIF
      IF(KSOFTE.EQ.0) RETURN
C
C  ****  Bielajew's randomly alternate hinge.
C
      IF(DE.GT.1.0D-3) THEN
        IF(RAND(2.0D0)*DST.GT.DS1) THEN
          XEL=LOG(E)
          XE=1.0D0+(XEL-DLEMP1)*DLFC
          KE=XE
          XEK=XE-KE
          IF(T1E(M,KE+1).GT.-78.3D0) THEN
            T1=EXP(T1P(M,KE)+(T1P(M,KE+1)-T1P(M,KE))*XEK)
            T2=EXP(T2P(M,KE)+(T2P(M,KE+1)-T2P(M,KE))*XEK)
          ELSE
            T1=0.0D0
            T2=0.0D0
          ENDIF
          IF(T1.LT.1.0D-35) RETURN
        ENDIF
      ENDIF
C  ****  1st and 2nd moments of the angular distribution.
      EMU1=0.5D0*(1.0D0-EXP(-DST*T1))
      EMU2=EMU1-(1.0D0-EXP(-DST*T2))/6.0D0
C  ****  Sampling from a two-bar histogram with these moments.
      PNUM=2.0D0*EMU1-3.0D0*EMU2
      PDEN=1.0D0-2.0D0*EMU1
      PB=PNUM/PDEN
      PA=PDEN+PB
      RND=RAND(3.0D0)
      IF(RND.LT.PA) THEN
        CDT=1.0D0-2.0D0*PB*(RND/PA)
      ELSE
        CDT=1.0D0-2.0D0*(PB+(1.0D0-PB)*((RND-PA)/(1.0D0-PA)))
      ENDIF
      DF=TWOPI*RAND(4.0D0)
      CALL DIRECT(CDT,DF,U,V,W)
      RETURN
C
C  ************  Hard event.
C
 3100 CONTINUE
      MODE=0
C  ****  A delta interaction (ICOL=7) occurs when the maximum
C        allowed step length is exceeded.
      IF(KDELTA.EQ.1) THEN
        ICOL=7
        DE=0.0D0
        RETURN
      ENDIF
C  ****  Interaction probabilities.
      STNOW=P(2)+P(3)+P(4)+P(5)+P(6)
C  ****  Random sampling of the interaction type.
      STS=MAX(STNOW,ST)*RAND(5.0D0)
      SS=P(2)
      IF(SS.GT.STS) GO TO 3200
      SS=SS+P(3)
      IF(SS.GT.STS) GO TO 3300
      SS=SS+P(4)
      IF(SS.GT.STS) GO TO 3400
      SS=SS+P(5)
      IF(SS.GT.STS) GO TO 3500
      SS=SS+P(6)
      IF(SS.GT.STS) GO TO 3600
      SS=SS+P(8)
      IF(SS.GT.STS) GO TO 3800
C  ****  A delta interaction (ICOL=7) may occur when the total
C        interaction probability per unit path length, ST, is
C        larger than STNOW.
      ICOL=7
      DE=0.0D0
      RETURN
C
C  ****  Hard elastic collision (ICOL=2).
C
 3200 ICOL=2
      IF(E.GE.PELMAX(M)) THEN
        TRNDC=RNDCP(M,KE)+(RNDCP(M,KE+1)-RNDCP(M,KE))*XEK
        TA=EXP(AP(M,KE)+(AP(M,KE+1)-AP(M,KE))*XEK)
        TB=BP(M,KE)+(BP(M,KE+1)-BP(M,KE))*XEK
        CALL EELa(TA,TB,TRNDC,RMU)
      ELSE
        TRNDC=RNDCPd(M,KE)+(RNDCPd(M,KE+1)-RNDCPd(M,KE))*XEK
        CALL PELd(TRNDC,RMU)  ! Uses the ELSEPA database.
      ENDIF
      CDT=1.0D0-2.0D0*RMU
      DF=TWOPI*RAND(6.0D0)
      CALL DIRECT(CDT,DF,U,V,W)
      DE=0.0D0
      RETURN
C
C  ****  Hard inelastic collision (ICOL=3).
C
 3300 ICOL=3
      DELTA=DEL(M,KE)+(DEL(M,KE+1)-DEL(M,KE))*XEK
      CALL PINa(E,DELTA,DE,EP,CDT,ES,CDTS,M,IOSC)
!cx:
      if (de.lt.eabs(1,m)) derest = de
C  ****  Scattering angles (primary particle).
      DF=TWOPI*RAND(7.0D0)
C  ****  Delta ray.
      IF(ES.GT.EABS(1,M)) THEN
        DFS=DF+PI
        US=U
        VS=V
        WS=W
        CALL DIRECT(CDTS,DFS,US,VS,WS)
        ILBA(1)=ILB(1)+1
        ILBA(2)=KPAR
        ILBA(3)=ICOL
        ILBA(4)=0
        ILBA(5)=ILB(5)
        CALL STORES(ES,X,Y,Z,US,VS,WS,WGHT,1,ILBA,0)
      ENDIF
C  ****  New energy and direction.
      IF(EP.GT.EABS(3,M)) THEN
        E=EP
        CALL DIRECT(CDT,DF,U,V,W)
        RETURN
      ENDIF
      DE=E+TREV
      E=0.0D0
C  ****  Annihilation at rest.
      CALL PANaR
      RETURN
C
C  ****  Hard bremsstrahlung emission (ICOL=4).
C
 3400 ICOL=4
      CALL EBRa(E,DE,M)
C  ****  Bremsstrahlung photon.
      IF(DE.GT.EABS(2,M)) THEN
        CALL EBRaA(E,DE,CDTS,M)
        DFS=TWOPI*RAND(8.0D0)
        US=U
        VS=V
        WS=W
        CALL DIRECT(CDTS,DFS,US,VS,WS)
        ILBA(1)=ILB(1)+1
        ILBA(2)=KPAR
        ILBA(3)=ICOL
        ILBA(4)=0
        ILBA(5)=ILB(5)
        CALL STORES(DE,X,Y,Z,US,VS,WS,WGHT,2,ILBA,0)
!cx:
      else
        derest = de
      ENDIF
C  ****  New energy.
      E=E-DE
      IF(E.GT.EABS(3,M)) RETURN
      DE=E+DE+TREV
      E=0.0D0
C  ****  Annihilation at rest.
      CALL PANaR
      RETURN
C
C  ****  Ionisation of an inner shell (ICOL=5) -does not affect the
C        primary positron.
C
 3500 ICOL=5
      DE=0.0D0
      CALL PSIa(IZA,ISA)
      IF(IZA.LT.1) RETURN
      ILBA(3)=ICOL
      CALL RELAX(IZA,ISA)
      RETURN
C
C  ****  Positron annihilation in flight (ICOL=6).
C
 3600 ICOL=6
      CALL PANa(E1,CDT1,E2,CDT2)
      DF=TWOPI*RAND(9.0D0)
      IF(E1.GT.EABS(2,M)) THEN
        US=U
        VS=V
        WS=W
        CALL DIRECT(CDT1,DF,US,VS,WS)
        ILBA(1)=ILB(1)+1
        ILBA(2)=KPAR
        ILBA(3)=ICOL
        ILBA(4)=0
        ILBA(5)=ILB(5)
        CALL STORES(E1,X,Y,Z,US,VS,WS,WGHT,2,ILBA,0)
      ENDIF
      IF(E2.GT.EABS(2,M)) THEN
        DF=DF+PI
        US=U
        VS=V
        WS=W
        CALL DIRECT(CDT2,DF,US,VS,WS)
        ILBA(1)=ILB(1)+1
        ILBA(2)=KPAR
        ILBA(3)=ICOL
        ILBA(4)=0
        ILBA(5)=ILB(5)
        CALL STORES(E2,X,Y,Z,US,VS,WS,WGHT,2,ILBA,0)
      ENDIF
      DE=E+TREV
      E=0.0D0
      RETURN
C
C  ****  Auxiliary fictitious mechanism (ICOL=8).
C
 3800 ICOL=8
      DE=0.0D0
      CALL PAUX
      RETURN
      END


!>>>> End Of File >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
